import pytest

from zapros import URL
from zapros._handlers._mock import (
    Mock,
    MockMiddleware,
    MockRouter,
)
from zapros._models import (
    Headers,
    Request,
    Response,
)
from zapros.matchers import (
    and_,
    header,
    host,
    json,
    method,
    not_,
    or_,
    path,
    query,
)


def test_method_matcher():
    matcher = method("GET")
    request = Request(
        URL("https://example.com/path"),
        "GET",
    )
    assert matcher.match(request)

    request_post = Request(
        URL("https://example.com/path"),
        "POST",
    )
    assert not matcher.match(request_post)


def test_method_matcher_case_insensitive():
    matcher = method("get")
    request = Request(
        URL("https://example.com/path"),
        "GET",
    )
    assert matcher.match(request)


def test_path_matcher():
    matcher = path("/health")
    request = Request(
        URL("https://example.com/health"),
        "GET",
    )
    assert matcher.match(request)

    request_other = Request(
        URL("https://example.com/status"),
        "GET",
    )
    assert not matcher.match(request_other)


def test_host_matcher():
    matcher = host("api.example.com")
    request = Request(
        URL("https://api.example.com/path"),
        "GET",
    )
    assert matcher.match(request)

    request_other = Request(
        URL("https://other.example.com/path"),
        "GET",
    )
    assert not matcher.match(request_other)


def test_header_matcher():
    matcher = header("x-api-key", "secret")
    headers = Headers({"x-api-key": "secret"})
    request = Request(
        URL("https://example.com/path"),
        "GET",
        headers=headers,
    )
    assert matcher.match(request)

    headers_wrong = Headers({"x-api-key": "wrong"})
    request_wrong = Request(
        URL("https://example.com/path"),
        "GET",
        headers=headers_wrong,
    )
    assert not matcher.match(request_wrong)


def test_query_matcher():
    matcher = query(page="2", limit="10")
    request = Request(
        URL("https://example.com/path?page=2&limit=10"),
        "GET",
    )
    assert matcher.match(request)

    request_missing = Request(
        URL("https://example.com/path?page=2"),
        "GET",
    )
    assert not matcher.match(request_missing)


def test_json_matcher():
    matcher = json(lambda data: data.get("name") == "test")
    request = Request(
        URL("https://example.com/path"),
        "POST",
        json={"name": "test"},
    )
    assert matcher.match(request)

    request_wrong = Request(
        URL("https://example.com/path"),
        "POST",
        json={"name": "other"},
    )
    assert not matcher.match(request_wrong)


def test_and_matcher():
    matcher = and_(method("GET"), path("/health"))
    request = Request(
        URL("https://example.com/health"),
        "GET",
    )
    assert matcher.match(request)

    request_wrong_method = Request(
        URL("https://example.com/health"),
        "POST",
    )
    assert not matcher.match(request_wrong_method)

    request_wrong_path = Request(
        URL("https://example.com/status"),
        "GET",
    )
    assert not matcher.match(request_wrong_path)


def test_or_matcher():
    matcher = or_(path("/health"), path("/status"))
    request_health = Request(
        URL("https://example.com/health"),
        "GET",
    )
    assert matcher.match(request_health)

    request_status = Request(
        URL("https://example.com/status"),
        "GET",
    )
    assert matcher.match(request_status)

    request_other = Request(
        URL("https://example.com/other"),
        "GET",
    )
    assert not matcher.match(request_other)


def test_not_matcher():
    matcher = not_(method("GET"))
    request_post = Request(
        URL("https://example.com/path"),
        "POST",
    )
    assert matcher.match(request_post)

    request_get = Request(
        URL("https://example.com/path"),
        "GET",
    )
    assert not matcher.match(request_get)


def test_mock_given_respond():
    mock = Mock.given(path("/health")).respond(
        Response(
            status=200,
            json={"status": "ok"},
        )
    )
    request = Request(
        URL("https://example.com/health"),
        "GET",
    )

    assert mock.matches(request)
    response = mock.handle(request)
    assert response.status == 200


def test_mock_and_chaining():
    mock = Mock.given(method("GET")).and_(path("/health")).respond(Response(status=200, text="OK"))
    request = Request(
        URL("https://example.com/health"),
        "GET",
    )
    assert mock.matches(request)

    request_wrong = Request(
        URL("https://example.com/health"),
        "POST",
    )
    assert not mock.matches(request_wrong)


def test_mock_respond_text():
    mock = Mock.given(path("/hello")).respond(Response(status=200, text="Hello World"))
    request = Request(
        URL("https://example.com/hello"),
        "GET",
    )
    response = mock.handle(request)

    assert response.status == 200
    assert response.headers.get("content-type") == "text/plain; charset=utf-8"
    assert response.headers.get("content-length") == "11"


def test_mock_respond_json():
    mock = Mock.given(path("/data")).respond(
        Response(
            status=200,
            json={"key": "value"},
        )
    )
    request = Request(
        URL("https://example.com/data"),
        "GET",
    )
    response = mock.handle(request)

    assert response.status == 200
    assert response.headers.get("content-type") == "application/json; charset=utf-8"


def test_mock_respond_with_headers():
    mock = Mock.given(path("/custom")).respond(
        Response(
            status=201,
            text="Created",
            headers={"x-custom": "header"},
        )
    )
    request = Request(
        URL("https://example.com/custom"),
        "GET",
    )
    response = mock.handle(request)

    assert response.status == 201
    assert response.headers.get("x-custom") == "header"


def test_mock_callback():
    from zapros._models import Response

    def custom_handler(
        req: Request,
    ) -> Response:
        status = 404 if req.url.pathname == "/notfound" else 200
        return Response(
            status=status,
            headers=Headers({}),
            content=None,
        )

    mock = Mock.given(method("GET")).callback(custom_handler)

    request_found = Request(
        URL("https://example.com/found"),
        "GET",
    )
    response_found = mock.handle(request_found)
    assert response_found.status == 200

    request_notfound = Request(
        URL("https://example.com/notfound"),
        "GET",
    )
    response_notfound = mock.handle(request_notfound)
    assert response_notfound.status == 404


def test_mock_expect():
    mock = Mock.given(path("/api")).respond(Response(status=200)).expect(2)
    request = Request(
        URL("https://example.com/api"),
        "GET",
    )

    mock.handle(request)
    mock.handle(request)

    mock.verify()


def test_mock_expect_failure():
    mock = Mock.given(path("/api")).respond(Response(status=200)).expect(2).name("API Mock")
    request = Request(
        URL("https://example.com/api"),
        "GET",
    )

    mock.handle(request)

    with pytest.raises(
        AssertionError,
        match="API Mock: expected 2 calls, got 1",
    ):
        mock.verify()


def test_mock_once():
    mock = Mock.given(path("/api")).respond(Response(status=200)).once()
    request = Request(
        URL("https://example.com/api"),
        "GET",
    )

    mock.handle(request)
    mock.verify()


def test_mock_never():
    mock = Mock.given(path("/api")).respond(Response(status=200)).never()
    mock.verify()


def test_mock_never_failure():
    mock = Mock.given(path("/api")).respond(Response(status=200)).never()
    request = Request(
        URL("https://example.com/api"),
        "GET",
    )

    mock.handle(request)

    with pytest.raises(
        AssertionError,
        match="expected 0 calls, got 1",
    ):
        mock.verify()


def test_mock_reset():
    mock = Mock.given(path("/api")).respond(Response(status=200)).once()
    request = Request(
        URL("https://example.com/api"),
        "GET",
    )

    mock.handle(request)
    mock.verify()

    mock.reset()

    with pytest.raises(AssertionError):
        mock.verify()


def test_router_add_and_dispatch():
    router = MockRouter()
    mock1 = Mock.given(path("/health")).respond(Response(status=200, text="OK"))
    mock2 = Mock.given(path("/status")).respond(
        Response(
            status=200,
            json={"status": "up"},
        )
    )

    router.add(mock1).add(mock2)

    request_health = Request(
        URL("https://example.com/health"),
        "GET",
    )
    response = router.dispatch(request_health)
    assert response is not None
    assert response.status == 200

    request_status = Request(
        URL("https://example.com/status"),
        "GET",
    )
    response = router.dispatch(request_status)
    assert response is not None
    assert response.status == 200


def test_router_dispatch_no_match():
    router = MockRouter()
    mock = Mock.given(path("/health")).respond(Response(status=200))
    router.add(mock)

    request = Request(
        URL("https://example.com/notfound"),
        "GET",
    )
    response = router.dispatch(request)
    assert response is None


def test_router_dispatch_first_match():
    router = MockRouter()
    mock1 = Mock.given(path("/api")).respond(Response(status=200, text="First"))
    mock2 = Mock.given(path("/api")).respond(Response(status=200, text="Second"))

    router.add(mock1).add(mock2)

    request = Request(
        URL("https://example.com/api"),
        "GET",
    )
    response = router.dispatch(request)
    assert response is not None
    assert response.status == 200


def test_router_verify():
    router = MockRouter()
    mock1 = Mock.given(path("/health")).respond(Response(status=200)).once()
    mock2 = Mock.given(path("/status")).respond(Response(status=200)).never()

    router.add(mock1).add(mock2)

    request = Request(
        URL("https://example.com/health"),
        "GET",
    )
    router.dispatch(request)

    router.verify()


def test_router_verify_failure():
    router = MockRouter()
    mock = Mock.given(path("/api")).respond(Response(status=200)).once()
    router.add(mock)

    with pytest.raises(AssertionError):
        router.verify()


def test_router_reset():
    router = MockRouter()
    mock = Mock.given(path("/api")).respond(Response(status=200)).once()
    router.add(mock)

    request = Request(
        URL("https://example.com/api"),
        "GET",
    )
    router.dispatch(request)
    router.verify()

    router.reset()

    with pytest.raises(AssertionError):
        router.verify()


def test_complex_matcher_combination():
    matcher = and_(
        method("POST"),
        or_(
            path("/users"),
            path("/accounts"),
        ),
        header(
            "content-type",
            "application/json",
        ),
    )

    request_users = Request(
        URL("https://example.com/users"),
        "POST",
        headers=Headers({"content-type": "application/json"}),
    )
    assert matcher.match(request_users)

    request_accounts = Request(
        URL("https://example.com/accounts"),
        "POST",
        headers=Headers({"content-type": "application/json"}),
    )
    assert matcher.match(request_accounts)

    request_wrong_path = Request(
        URL("https://example.com/other"),
        "POST",
        headers=Headers({"content-type": "application/json"}),
    )
    assert not matcher.match(request_wrong_path)

    request_wrong_method = Request(
        URL("https://example.com/users"),
        "GET",
        headers=Headers({"content-type": "application/json"}),
    )
    assert not matcher.match(request_wrong_method)


def test_helper_functions():
    m = method("GET")
    request = Request(
        URL("https://example.com/path"),
        "GET",
    )
    assert m.match(request)

    p = path("/health")
    request = Request(
        URL("https://example.com/health"),
        "GET",
    )
    assert p.match(request)

    h = host("api.example.com")
    request = Request(
        URL("https://api.example.com/path"),
        "GET",
    )
    assert h.match(request)


def test_chaining_with_helper_functions():
    matcher = path("/api").method("POST").host("api.example.com")

    request = Request(
        URL("https://api.example.com/api"),
        "POST",
    )
    assert matcher.match(request)

    request_wrong_method = Request(
        URL("https://api.example.com/api"),
        "GET",
    )
    assert not matcher.match(request_wrong_method)

    request_wrong_host = Request(
        URL("https://other.example.com/api"),
        "POST",
    )
    assert not matcher.match(request_wrong_host)


def test_chaining_with_query_and_header():
    matcher = (
        path("/search")
        .query(q="test", page="1")
        .header(
            "authorization",
            "Bearer token",
        )
    )

    request = Request(
        URL("https://example.com/search?q=test&page=1"),
        "GET",
        headers=Headers({"authorization": "Bearer token"}),
    )
    assert matcher.match(request)

    request_wrong_query = Request(
        URL("https://example.com/search?q=test"),
        "GET",
        headers=Headers({"authorization": "Bearer token"}),
    )
    assert not matcher.match(request_wrong_query)


def test_chaining_with_and_or():
    matcher = method("GET").and_(
        or_(
            path("/health"),
            path("/status"),
        )
    )

    request_health = Request(
        URL("https://example.com/health"),
        "GET",
    )
    assert matcher.match(request_health)

    request_status = Request(
        URL("https://example.com/status"),
        "GET",
    )
    assert matcher.match(request_status)

    request_other = Request(
        URL("https://example.com/other"),
        "GET",
    )
    assert not matcher.match(request_other)


def test_mock_with_chainedmatchers():
    mock = (
        Mock.given(
            path("/api/users")
            .method("POST")
            .header(
                "content-type",
                "application/json",
            )
        )
        .respond(Response(status=201, json={"id": 123}))
        .once()
    )

    request = Request(
        URL("https://example.com/api/users"),
        "POST",
        headers=Headers({"content-type": "application/json"}),
    )
    assert mock.matches(request)
    response = mock.handle(request)
    assert response.status == 201
    mock.verify()


def test_router_with_chainedmatchers():
    router = MockRouter()
    mock1 = Mock.given(method("GET").path("/health")).respond(Response(status=200, text="OK"))
    mock2 = Mock.given(method("GET").path("/status")).respond(
        Response(
            status=200,
            json={"status": "up"},
        )
    )

    router.add(mock1).add(mock2)

    request_health = Request(
        URL("https://example.com/health"),
        "GET",
    )
    response = router.dispatch(request_health)
    assert response is not None
    assert response.status == 200


def test_given_standalone_function():
    mock = Mock.given(path("/api").method("POST")).respond(
        Response(
            status=201,
            json={"created": True},
        )
    )

    request = Request(
        URL("https://example.com/api"),
        "POST",
    )
    assert mock.matches(request)
    response = mock.handle(request)
    assert response.status == 201


def test_mount_method():
    router = MockRouter()
    Mock.given(path("/health")).respond(Response(status=200, text="OK")).mount(router)

    request = Request(
        URL("https://example.com/health"),
        "GET",
    )
    response = router.dispatch(request)
    assert response is not None
    assert response.status == 200


def test_fluent_api_complete():
    router = MockRouter()

    Mock.given(path("/api/users").method("GET")).respond(Response(status=200, json=[])).once().mount(router)

    Mock.given(path("/api/users").method("POST")).respond(Response(status=201, json={"id": 1})).once().mount(router)

    request_get = Request(
        URL("https://example.com/api/users"),
        "GET",
    )
    response_get = router.dispatch(request_get)
    assert response_get is not None
    assert response_get.status == 200

    request_post = Request(
        URL("https://example.com/api/users"),
        "POST",
    )
    response_post = router.dispatch(request_post)
    assert response_post is not None
    assert response_post.status == 201

    router.verify()


def test_mock_callback_exception_instance():
    mock = Mock.given(path("/error")).callback(ValueError("something went wrong"))
    request = Request(
        URL("https://example.com/error"),
        "GET",
    )

    with pytest.raises(
        ValueError,
        match="something went wrong",
    ):
        mock.handle(request)


def test_mock_callback_exception_class():
    mock = Mock.given(path("/error")).callback(ValueError)
    request = Request(
        URL("https://example.com/error"),
        "GET",
    )

    with pytest.raises(ValueError):
        mock.handle(request)


@pytest.mark.asyncio
async def test_async_mock_handler_dispatch():
    router = MockRouter()
    Mock.given(path("/health")).respond(Response(status=200, text="OK")).mount(router)
    Mock.given(path("/api").method("POST")).respond(
        Response(
            status=201,
            json={"created": True},
        )
    ).mount(router)

    handler = MockMiddleware(router)

    request_health = Request(
        URL("https://example.com/health"),
        "GET",
    )
    response = await handler.ahandle(request_health)
    assert response.status == 200

    request_api = Request(
        URL("https://example.com/api"),
        "POST",
    )
    response = await handler.ahandle(request_api)
    assert response.status == 201


@pytest.mark.asyncio
async def test_async_mock_handler_no_match():
    router = MockRouter()
    Mock.given(path("/health")).respond(Response(status=200, text="OK")).mount(router)

    handler = MockMiddleware(router)

    request = Request(
        URL("https://example.com/notfound"),
        "GET",
    )
    with pytest.raises(
        ValueError,
        match="No mock matched request: GET /notfound",
    ):
        await handler.ahandle(request)


@pytest.mark.asyncio
async def test_async_mock_handler_with_fallback():
    fallback_router = MockRouter()
    Mock().callback(
        lambda request: Response(
            status=404,
            headers=Headers({}),
            content=None,
        )
    ).mount(fallback_router)

    router = MockRouter()
    Mock.given(path("/health")).respond(Response(status=200, text="OK")).mount(router)

    fallback = MockMiddleware(fallback_router)
    handler = MockMiddleware(router, next_handler=fallback)

    request = Request(
        URL("https://example.com/notfound"),
        "GET",
    )
    response = await handler.ahandle(request)
    assert response.status == 404



from __future__ import annotations

import asyncio
import random
import time
from abc import abstractmethod
from typing import (
    Protocol,
    cast,
    runtime_checkable,
)

import typing_extensions

from zapros._handlers._common import (
    ensure_async_handler,
    ensure_sync_handler,
)

from .._errors import (
    ConnectionError,
    TimeoutError,
)
from .._models import Request, Response
from ._async_base import (
    AsyncBaseHandler,
    AsyncBaseMiddleware,
)
from ._sync_base import BaseHandler

DEFAULT_RETRY_STATUS_CODES = frozenset(
    {
        429,
        500,
        502,
        503,
        504,
    }
)
SAFE_RETRY_METHODS = frozenset(
    {
        "GET",
        "HEAD",
        "PUT",
        "DELETE",
        "OPTIONS",
        "TRACE",
    }
)
DEFAULT_MAX_ATTEMPTS = 4
DEFAULT_BACKOFF_FACTOR = 0.5
DEFAULT_BACKOFF_MAX = 60.0
DEFAULT_BACKOFF_JITTER = 1.0


@runtime_checkable
class RetryPolicy(Protocol):
    @abstractmethod
    def should_retry(
        self,
        *,
        request: Request,
        response: Response | None,
        error: Exception | None,
        attempt: int,
    ) -> bool:
        raise NotImplementedError()


class DefaultRetryPolicy:
    def __init__(
        self,
        *,
        retry_status_codes: frozenset[int] | None = None,
        safe_methods: frozenset[str] | None = None,
    ) -> None:
        self._retry_status_codes = retry_status_codes or DEFAULT_RETRY_STATUS_CODES
        self._safe_methods = safe_methods or SAFE_RETRY_METHODS

    def should_retry(
        self,
        *,
        request: Request,
        response: Response | None,
        error: Exception | None,
        attempt: int,
    ) -> bool:
        if error is not None:
            if self._is_pre_transmission_error(error):
                return True
            if request.method not in self._safe_methods:
                return False
            if request.is_replayable() and self._is_retryable_exception(error):
                return True
            return False

        if response is not None:
            if request.method not in self._safe_methods:
                return False
            if not request.is_replayable():
                return False
            return response.status in self._retry_status_codes

        return False

    def _is_pre_transmission_error(self, exc: Exception) -> bool:
        exc_name = type(exc).__name__
        pre_transmission_patterns = (
            "ConnectionError",
            "ConnectionRefusedError",
            "ConnectTimeout",
            "ConnectError",
            "DNSError",
            "NameResolutionError",
            "SSLError",
            "CertificateError",
        )
        return any(pattern in exc_name for pattern in pre_transmission_patterns)

    def _is_retryable_exception(self, exc: Exception) -> bool:
        return isinstance(
            exc,
            (
                TimeoutError,
                ConnectionError,
            ),
        )


def _calculate_backoff(
    attempt: int,
    backoff_factor: float,
    backoff_max: float,
    backoff_jitter: float,
) -> float:
    base_delay = backoff_factor * (2**attempt)
    delay = min(base_delay, backoff_max)

    if backoff_jitter > 0:
        jitter_range = delay * backoff_jitter
        delay = delay - jitter_range + (random.random() * 2 * jitter_range)
        delay = max(0.0, delay)

    return delay


class RetryMiddleware(AsyncBaseMiddleware, BaseHandler):
    def __init__(
        self,
        next_handler: AsyncBaseHandler | BaseHandler,
        *,
        policy: RetryPolicy | None = None,
        max_attempts: int = DEFAULT_MAX_ATTEMPTS,
        backoff_factor: float = DEFAULT_BACKOFF_FACTOR,
        backoff_max: float = DEFAULT_BACKOFF_MAX,
        backoff_jitter: float = DEFAULT_BACKOFF_JITTER,
    ) -> None:
        self.next = cast(BaseHandler, next_handler)
        self.async_next = cast(
            AsyncBaseMiddleware,
            next_handler,
        )
        self._policy = policy if policy is not None else DefaultRetryPolicy()
        self._max_attempts = max_attempts
        self._backoff_factor = backoff_factor
        self._backoff_max = backoff_max
        self._backoff_jitter = backoff_jitter

    async def ahandle(self, request: Request) -> Response:
        handler = ensure_async_handler(self.async_next)

        attempt = 0
        while attempt < self._max_attempts:
            try:
                response = await handler.ahandle(request)

                if not self._policy.should_retry(
                    request=request,
                    response=response,
                    error=None,
                    attempt=attempt,
                ):
                    return response

                if attempt + 1 >= self._max_attempts:
                    return response

                await response.aclose()

                wait = _calculate_backoff(
                    attempt,
                    self._backoff_factor,
                    self._backoff_max,
                    self._backoff_jitter,
                )
                if wait > 0:
                    await asyncio.sleep(wait)

                attempt += 1

            except Exception as exc:
                if not self._policy.should_retry(
                    request=request,
                    response=None,
                    error=exc,
                    attempt=attempt,
                ):
                    raise

                if attempt + 1 >= self._max_attempts:
                    raise

                wait = _calculate_backoff(
                    attempt,
                    self._backoff_factor,
                    self._backoff_max,
                    self._backoff_jitter,
                )
                if wait > 0:
                    await asyncio.sleep(wait)

                attempt += 1

        raise RuntimeError("Unreachable code")

    def handle(self, request: Request) -> Response:
        handler = ensure_sync_handler(self.next)

        attempt = 0
        while attempt < self._max_attempts:
            try:
                response = handler.handle(request)

                if not self._policy.should_retry(
                    request=request,
                    response=response,
                    error=None,
                    attempt=attempt,
                ):
                    return response

                if attempt + 1 >= self._max_attempts:
                    return response

                response.close()

                wait = _calculate_backoff(
                    attempt,
                    self._backoff_factor,
                    self._backoff_max,
                    self._backoff_jitter,
                )
                if wait > 0:
                    time.sleep(wait)

                attempt += 1

            except Exception as exc:
                if not self._policy.should_retry(
                    request=request,
                    response=None,
                    error=exc,
                    attempt=attempt,
                ):
                    raise

                if attempt + 1 >= self._max_attempts:
                    raise

                wait = _calculate_backoff(
                    attempt,
                    self._backoff_factor,
                    self._backoff_max,
                    self._backoff_jitter,
                )
                if wait > 0:
                    time.sleep(wait)

                attempt += 1

        raise RuntimeError("Unreachable code")


@typing_extensions.deprecated(
    "RetryHandler is deprecated, use RetryMiddleware instead. "
    "The name 'Handler' was misleading as this is a middleware, not a terminal handler."
)
class RetryHandler(RetryMiddleware):
    pass



import base64
import sys
from contextlib import (
    AbstractContextManager,
    contextmanager,
)
from typing import (
    TYPE_CHECKING,
    Any,
    Iterable,
    Callable,
    Mapping,
    Sequence,
    Union,
    overload,
)

from pywhatwgurl import (
    URL,
    URLSearchParams,
)
from typing_extensions import Self

from zapros._handlers._sync_base import (
    BaseHandler,
    BaseMiddleware,
)

from ._models import (
    ClosableStream,
    Stream,
    Headers,
    Request,
    RequestContext,
    Response,
)

if TYPE_CHECKING:
    from zapros import BaseHandler

    from ._multipart import Multipart


class Client:
    @overload
    def __init__(
        self,
        handler: Union[None, "BaseHandler"] = None,
        *,
        default_headers: Mapping[str, str] | Headers | None = None,
        default_params: Union[
            str,
            Iterable[Sequence[str]],
            Mapping[str, Union[str, Sequence[str]]],
            URLSearchParams,
        ]
        | None = None,
    ) -> None: ...

    @overload
    def __init__(
        self,
        handler: Union[None, "BaseHandler"] = None,
        *,
        default_headers: Mapping[str, str] | Headers | None = None,
        default_params: Union[
            str,
            Iterable[Sequence[str]],
            Mapping[str, Union[str, Sequence[str]]],
            URLSearchParams,
        ]
        | None = None,
        auth: str | tuple[str, str],
    ) -> None: ...

    def __init__(
        self,
        handler: Union[None, "BaseHandler"] = None,
        *,
        default_headers: Mapping[str, str] | Headers | None = None,
        default_params: Union[
            str,
            Iterable[Sequence[str]],
            Mapping[str, Union[str, Sequence[str]]],
            URLSearchParams,
        ]
        | None = None,
        auth: str | tuple[str, str] | None = None,
    ) -> None:
        from zapros import (
            StdNetworkHandler,
        )

        if sys.platform == "emscripten" and handler is None:
            from zapros import (
                PyodideHandler,
            )

            self.handler = PyodideHandler()
        else:
            self.handler = handler if handler is not None else StdNetworkHandler()
        self.default_headers: Headers = (
            default_headers if isinstance(default_headers, Headers) else Headers(default_headers)
        )
        self.default_params = URLSearchParams(default_params)
        self.auth = auth

    def _merge_url(
        self,
        url: str | URL,
        params: Union[
            str,
            Iterable[Sequence[str]],
            Mapping[str, Union[str, Sequence[str]]],
            URLSearchParams,
        ]
        | None = None,
    ) -> URL:
        url_obj = url if isinstance(url, URL) else URL(url)
        url_obj.search = URLSearchParams(
            {
                **self.default_params,
                **url_obj.search_params,
                **(URLSearchParams(params)),
            }
        ).to_string()
        return url_obj

    def _merge_headers(
        self,
        headers: Mapping[str, str] | Headers | None,
        auth: str | tuple[str, str] | None = None,
    ) -> Headers:
        merged = dict(self.default_headers.items())

        request_auth = auth if auth is not None else self.auth

        if request_auth is not None:
            if isinstance(request_auth, str):
                merged["Authorization"] = f"Bearer {request_auth}"
            else:
                username, password = request_auth
                credentials = base64.b64encode(f"{username}:{password}".encode()).decode()
                merged["Authorization"] = f"Basic {credentials}"

        if headers is not None:
            merged.update(headers.items() if isinstance(headers, Headers) else headers)
        return Headers(merged)

    @overload
    def request(
        self,
        method: str,
        url: str | URL,
        *,
        headers: Mapping[str, str] | Headers | None = None,
        params: Union[
            str,
            Iterable[Sequence[str]],
            Mapping[str, Union[str, Sequence[str]]],
            URLSearchParams,
        ]
        | None = None,
        auth: str | tuple[str, str] | None = None,
        context: RequestContext | None = None,
        json: Any,
    ) -> Response: ...

    @overload
    def request(
        self,
        method: str,
        url: str | URL,
        *,
        headers: Mapping[str, str] | Headers | None = None,
        params: Union[
            str,
            Iterable[Sequence[str]],
            Mapping[str, Union[str, Sequence[str]]],
            URLSearchParams,
        ]
        | None = None,
        auth: str | tuple[str, str] | None = None,
        context: RequestContext | None = None,
        form: Union[
            str,
            Iterable[Sequence[str]],
            Mapping[str, Union[str, Sequence[str]]],
            URLSearchParams,
        ],
    ) -> Response: ...

    @overload
    def request(
        self,
        method: str,
        url: str | URL,
        *,
        headers: Mapping[str, str] | Headers | None = None,
        params: Union[
            str,
            Iterable[Sequence[str]],
            Mapping[str, Union[str, Sequence[str]]],
            URLSearchParams,
        ]
        | None = None,
        auth: str | tuple[str, str] | None = None,
        context: RequestContext | None = None,
        body: bytes | Stream,
    ) -> Response: ...

    @overload
    def request(
        self,
        method: str,
        url: str | URL,
        *,
        headers: Mapping[str, str] | Headers | None = None,
        params: Union[
            str,
            Iterable[Sequence[str]],
            Mapping[str, Union[str, Sequence[str]]],
            URLSearchParams,
        ]
        | None = None,
        auth: str | tuple[str, str] | None = None,
        context: RequestContext | None = None,
        multipart: "Multipart",
    ) -> Response: ...

    @overload
    def request(
        self,
        method: str,
        url: str | URL,
        *,
        headers: Mapping[str, str] | Headers | None = None,
        params: Union[
            str,
            Iterable[Sequence[str]],
            Mapping[str, Union[str, Sequence[str]]],
            URLSearchParams,
        ]
        | None = None,
        auth: str | tuple[str, str] | None = None,
        context: RequestContext | None = None,
    ) -> Response: ...

    def request(
        self,
        method: str,
        url: str | URL,
        *,
        headers: Mapping[str, str] | Headers | None = None,
        params: Union[
            str,
            Iterable[Sequence[str]],
            Mapping[str, Union[str, Sequence[str]]],
            URLSearchParams,
        ]
        | None = None,
        auth: str | tuple[str, str] | None = None,
        context: RequestContext | None = None,
        json: Any | None = None,
        form: Union[
            str,
            Iterable[Sequence[str]],
            Mapping[str, Union[str, Sequence[str]]],
            URLSearchParams,
        ]
        | None = None,
        body: bytes | Stream | None = None,
        multipart: "Multipart | None" = None,
    ) -> Response:
        url_obj = self._merge_url(url, params)

        merged_headers = self._merge_headers(headers, auth=auth)

        if json is not None:
            request = Request(
                method=method,
                url=url_obj,
                headers=merged_headers,
                context=context,
                json=json,
            )
        elif form is not None:
            request = Request(
                method=method,
                url=url_obj,
                headers=merged_headers,
                context=context,
                form=form,
            )
        elif multipart is not None:
            request = Request(
                method=method,
                url=url_obj,
                headers=merged_headers,
                context=context,
                multipart=multipart,
            )
        elif body is not None:
            request = Request(
                method=method,
                url=url_obj,
                headers=merged_headers,
                context=context,
                body=body,
            )
        else:
            request = Request(
                method=method,
                url=url_obj,
                headers=merged_headers,
                context=context,
            )

        try:
            response = self.handler.handle(request=request)

            try:
                response.read()
            finally:
                response.close()
        finally:
            if isinstance(request.body, ClosableStream):
                request.body.close()

        return response

    def get(
        self,
        url: str | URL,
        *,
        headers: Mapping[str, str] | Headers | None = None,
        params: Union[
            str,
            Iterable[Sequence[str]],
            Mapping[str, Union[str, Sequence[str]]],
            URLSearchParams,
        ]
        | None = None,
        auth: str | tuple[str, str] | None = None,
        context: RequestContext | None = None,
    ) -> Response:
        return self.request(
            "GET",
            url,
            headers=headers,
            params=params,
            auth=auth,
            context=context,
        )

    @overload
    def post(
        self,
        url: str | URL,
        *,
        headers: Mapping[str, str] | Headers | None = None,
        params: Union[
            str,
            Iterable[Sequence[str]],
            Mapping[str, Union[str, Sequence[str]]],
            URLSearchParams,
        ]
        | None = None,
        auth: str | tuple[str, str] | None = None,
        context: RequestContext | None = None,
        json: Any,
    ) -> Response: ...

    @overload
    def post(
        self,
        url: str | URL,
        *,
        headers: Mapping[str, str] | Headers | None = None,
        params: Union[
            str,
            Iterable[Sequence[str]],
            Mapping[str, Union[str, Sequence[str]]],
            URLSearchParams,
        ]
        | None = None,
        auth: str | tuple[str, str] | None = None,
        context: RequestContext | None = None,
        form: Union[
            str,
            Iterable[Sequence[str]],
            Mapping[str, Union[str, Sequence[str]]],
            URLSearchParams,
        ],
    ) -> Response: ...

    @overload
    def post(
        self,
        url: str | URL,
        *,
        headers: Mapping[str, str] | Headers | None = None,
        params: Union[
            str,
            Iterable[Sequence[str]],
            Mapping[str, Union[str, Sequence[str]]],
            URLSearchParams,
        ]
        | None = None,
        auth: str | tuple[str, str] | None = None,
        context: RequestContext | None = None,
        body: bytes | Stream,
    ) -> Response: ...

    @overload
    def post(
        self,
        url: str | URL,
        *,
        headers: Mapping[str, str] | Headers | None = None,
        params: Union[
            str,
            Iterable[Sequence[str]],
            Mapping[str, Union[str, Sequence[str]]],
            URLSearchParams,
        ]
        | None = None,
        auth: str | tuple[str, str] | None = None,
        context: RequestContext | None = None,
        multipart: "Multipart",
    ) -> Response: ...

    @overload
    def post(
        self,
        url: str | URL,
        *,
        headers: Mapping[str, str] | Headers | None = None,
        params: Union[
            str,
            Iterable[Sequence[str]],
            Mapping[str, Union[str, Sequence[str]]],
            URLSearchParams,
        ]
        | None = None,
        auth: str | tuple[str, str] | None = None,
        context: RequestContext | None = None,
    ) -> Response: ...

    def post(
        self,
        url: str | URL,
        *,
        headers: Mapping[str, str] | Headers | None = None,
        params: Union[
            str,
            Iterable[Sequence[str]],
            Mapping[str, Union[str, Sequence[str]]],
            URLSearchParams,
        ]
        | None = None,
        auth: str | tuple[str, str] | None = None,
        context: RequestContext | None = None,
        json: Any | None = None,
        form: Union[
            str,
            Iterable[Sequence[str]],
            Mapping[str, Union[str, Sequence[str]]],
            URLSearchParams,
        ]
        | None = None,
        body: bytes | Stream | None = None,
        multipart: "Multipart | None" = None,
    ) -> Response:
        return self.request(  # type: ignore[call-overload]
            "POST",
            url,
            headers=headers,
            params=params,
            auth=auth,
            context=context,
            json=json,
            form=form,
            body=body,
            multipart=multipart,
        )

    @overload
    def put(
        self,
        url: str | URL,
        *,
        headers: Mapping[str, str] | Headers | None = None,
        params: Union[
            str,
            Iterable[Sequence[str]],
            Mapping[str, Union[str, Sequence[str]]],
            URLSearchParams,
        ]
        | None = None,
        auth: str | tuple[str, str] | None = None,
        context: RequestContext | None = None,
        json: Any,
    ) -> Response: ...

    @overload
    def put(
        self,
        url: str | URL,
        *,
        headers: Mapping[str, str] | Headers | None = None,
        params: Union[
            str,
            Iterable[Sequence[str]],
            Mapping[str, Union[str, Sequence[str]]],
            URLSearchParams,
        ]
        | None = None,
        auth: str | tuple[str, str] | None = None,
        context: RequestContext | None = None,
        form: Union[
            str,
            Iterable[Sequence[str]],
            Mapping[str, Union[str, Sequence[str]]],
            URLSearchParams,
        ],
    ) -> Response: ...

    @overload
    def put(
        self,
        url: str | URL,
        *,
        headers: Mapping[str, str] | Headers | None = None,
        params: Union[
            str,
            Iterable[Sequence[str]],
            Mapping[str, Union[str, Sequence[str]]],
            URLSearchParams,
        ]
        | None = None,
        auth: str | tuple[str, str] | None = None,
        context: RequestContext | None = None,
        body: bytes | Stream,
    ) -> Response: ...

    @overload
    def put(
        self,
        url: str | URL,
        *,
        headers: Mapping[str, str] | Headers | None = None,
        params: Union[
            str,
            Iterable[Sequence[str]],
            Mapping[str, Union[str, Sequence[str]]],
            URLSearchParams,
        ]
        | None = None,
        auth: str | tuple[str, str] | None = None,
        context: RequestContext | None = None,
        multipart: "Multipart",
    ) -> Response: ...

    @overload
    def put(
        self,
        url: str | URL,
        *,
        headers: Mapping[str, str] | Headers | None = None,
        params: Union[
            str,
            Iterable[Sequence[str]],
            Mapping[str, Union[str, Sequence[str]]],
            URLSearchParams,
        ]
        | None = None,
        auth: str | tuple[str, str] | None = None,
        context: RequestContext | None = None,
    ) -> Response: ...

    def put(
        self,
        url: str | URL,
        *,
        headers: Mapping[str, str] | Headers | None = None,
        params: Union[
            str,
            Iterable[Sequence[str]],
            Mapping[str, Union[str, Sequence[str]]],
            URLSearchParams,
        ]
        | None = None,
        auth: str | tuple[str, str] | None = None,
        context: RequestContext | None = None,
        json: Any | None = None,
        form: Union[
            str,
            Iterable[Sequence[str]],
            Mapping[str, Union[str, Sequence[str]]],
            URLSearchParams,
        ]
        | None = None,
        body: bytes | Stream | None = None,
        multipart: "Multipart | None" = None,
    ) -> Response:
        return self.request(  # type: ignore[call-overload]
            "PUT",
            url,
            headers=headers,
            params=params,
            auth=auth,
            context=context,
            json=json,
            form=form,
            body=body,
            multipart=multipart,
        )

    @overload
    def patch(
        self,
        url: str | URL,
        *,
        headers: Mapping[str, str] | Headers | None = None,
        params: Union[
            str,
            Iterable[Sequence[str]],
            Mapping[str, Union[str, Sequence[str]]],
            URLSearchParams,
        ]
        | None = None,
        auth: str | tuple[str, str] | None = None,
        context: RequestContext | None = None,
        json: Any,
    ) -> Response: ...

    @overload
    def patch(
        self,
        url: str | URL,
        *,
        headers: Mapping[str, str] | Headers | None = None,
        params: Union[
            str,
            Iterable[Sequence[str]],
            Mapping[str, Union[str, Sequence[str]]],
            URLSearchParams,
        ]
        | None = None,
        auth: str | tuple[str, str] | None = None,
        context: RequestContext | None = None,
        form: Union[
            str,
            Iterable[Sequence[str]],
            Mapping[str, Union[str, Sequence[str]]],
            URLSearchParams,
        ],
    ) -> Response: ...

    @overload
    def patch(
        self,
        url: str | URL,
        *,
        headers: Mapping[str, str] | Headers | None = None,
        params: Union[
            str,
            Iterable[Sequence[str]],
            Mapping[str, Union[str, Sequence[str]]],
            URLSearchParams,
        ]
        | None = None,
        auth: str | tuple[str, str] | None = None,
        context: RequestContext | None = None,
        body: bytes | Stream,
    ) -> Response: ...

    @overload
    def patch(
        self,
        url: str | URL,
        *,
        headers: Mapping[str, str] | Headers | None = None,
        params: Union[
            str,
            Iterable[Sequence[str]],
            Mapping[str, Union[str, Sequence[str]]],
            URLSearchParams,
        ]
        | None = None,
        auth: str | tuple[str, str] | None = None,
        context: RequestContext | None = None,
        multipart: "Multipart",
    ) -> Response: ...

    @overload
    def patch(
        self,
        url: str | URL,
        *,
        headers: Mapping[str, str] | Headers | None = None,
        params: Union[
            str,
            Iterable[Sequence[str]],
            Mapping[str, Union[str, Sequence[str]]],
            URLSearchParams,
        ]
        | None = None,
        auth: str | tuple[str, str] | None = None,
        context: RequestContext | None = None,
    ) -> Response: ...

    def patch(
        self,
        url: str | URL,
        *,
        headers: Mapping[str, str] | Headers | None = None,
        params: Union[
            str,
            Iterable[Sequence[str]],
            Mapping[str, Union[str, Sequence[str]]],
            URLSearchParams,
        ]
        | None = None,
        auth: str | tuple[str, str] | None = None,
        context: RequestContext | None = None,
        json: Any | None = None,
        form: Union[
            str,
            Iterable[Sequence[str]],
            Mapping[str, Union[str, Sequence[str]]],
            URLSearchParams,
        ]
        | None = None,
        body: bytes | Stream | None = None,
        multipart: "Multipart | None" = None,
    ) -> Response:
        return self.request(  # type: ignore[call-overload]
            "PATCH",
            url,
            headers=headers,
            params=params,
            auth=auth,
            context=context,
            json=json,
            form=form,
            body=body,
            multipart=multipart,
        )

    @overload
    def delete(
        self,
        url: str | URL,
        *,
        headers: Mapping[str, str] | Headers | None = None,
        params: Union[
            str,
            Iterable[Sequence[str]],
            Mapping[str, Union[str, Sequence[str]]],
            URLSearchParams,
        ]
        | None = None,
        auth: str | tuple[str, str] | None = None,
        context: RequestContext | None = None,
        json: Any,
    ) -> Response: ...

    @overload
    def delete(
        self,
        url: str | URL,
        *,
        headers: Mapping[str, str] | Headers | None = None,
        params: Union[
            str,
            Iterable[Sequence[str]],
            Mapping[str, Union[str, Sequence[str]]],
            URLSearchParams,
        ]
        | None = None,
        auth: str | tuple[str, str] | None = None,
        context: RequestContext | None = None,
        form: Union[
            str,
            Iterable[Sequence[str]],
            Mapping[str, Union[str, Sequence[str]]],
            URLSearchParams,
        ],
    ) -> Response: ...

    @overload
    def delete(
        self,
        url: str | URL,
        *,
        headers: Mapping[str, str] | Headers | None = None,
        params: Union[
            str,
            Iterable[Sequence[str]],
            Mapping[str, Union[str, Sequence[str]]],
            URLSearchParams,
        ]
        | None = None,
        auth: str | tuple[str, str] | None = None,
        context: RequestContext | None = None,
        body: bytes | Stream,
    ) -> Response: ...

    @overload
    def delete(
        self,
        url: str | URL,
        *,
        headers: Mapping[str, str] | Headers | None = None,
        params: Union[
            str,
            Iterable[Sequence[str]],
            Mapping[str, Union[str, Sequence[str]]],
            URLSearchParams,
        ]
        | None = None,
        auth: str | tuple[str, str] | None = None,
        context: RequestContext | None = None,
        multipart: "Multipart",
    ) -> Response: ...

    @overload
    def delete(
        self,
        url: str | URL,
        *,
        headers: Mapping[str, str] | Headers | None = None,
        params: Union[
            str,
            Iterable[Sequence[str]],
            Mapping[str, Union[str, Sequence[str]]],
            URLSearchParams,
        ]
        | None = None,
        auth: str | tuple[str, str] | None = None,
        context: RequestContext | None = None,
    ) -> Response: ...

    def delete(
        self,
        url: str | URL,
        *,
        headers: Mapping[str, str] | Headers | None = None,
        params: Union[
            str,
            Iterable[Sequence[str]],
            Mapping[str, Union[str, Sequence[str]]],
            URLSearchParams,
        ]
        | None = None,
        auth: str | tuple[str, str] | None = None,
        context: RequestContext | None = None,
        json: Any | None = None,
        form: Union[
            str,
            Iterable[Sequence[str]],
            Mapping[str, Union[str, Sequence[str]]],
            URLSearchParams,
        ]
        | None = None,
        body: bytes | Stream | None = None,
        multipart: "Multipart | None" = None,
    ) -> Response:
        return self.request(  # type: ignore[call-overload]
            "DELETE",
            url,
            headers=headers,
            params=params,
            auth=auth,
            context=context,
            json=json,
            form=form,
            body=body,
            multipart=multipart,
        )

    def head(
        self,
        url: str | URL,
        *,
        headers: Mapping[str, str] | Headers | None = None,
        params: Union[
            str,
            Iterable[Sequence[str]],
            Mapping[str, Union[str, Sequence[str]]],
            URLSearchParams,
        ]
        | None = None,
        auth: str | tuple[str, str] | None = None,
        context: RequestContext | None = None,
    ) -> Response:
        return self.request(
            "HEAD",
            url,
            headers=headers,
            params=params,
            auth=auth,
            context=context,
        )

    @overload
    def options(
        self,
        url: str | URL,
        *,
        headers: Mapping[str, str] | Headers | None = None,
        params: Union[
            str,
            Iterable[Sequence[str]],
            Mapping[str, Union[str, Sequence[str]]],
            URLSearchParams,
        ]
        | None = None,
        auth: str | tuple[str, str] | None = None,
        context: RequestContext | None = None,
        json: Any,
    ) -> Response: ...

    @overload
    def options(
        self,
        url: str | URL,
        *,
        headers: Mapping[str, str] | Headers | None = None,
        params: Union[
            str,
            Iterable[Sequence[str]],
            Mapping[str, Union[str, Sequence[str]]],
            URLSearchParams,
        ]
        | None = None,
        auth: str | tuple[str, str] | None = None,
        context: RequestContext | None = None,
        form: Union[
            str,
            Iterable[Sequence[str]],
            Mapping[str, Union[str, Sequence[str]]],
            URLSearchParams,
        ],
    ) -> Response: ...

    @overload
    def options(
        self,
        url: str | URL,
        *,
        headers: Mapping[str, str] | Headers | None = None,
        params: Union[
            str,
            Iterable[Sequence[str]],
            Mapping[str, Union[str, Sequence[str]]],
            URLSearchParams,
        ]
        | None = None,
        auth: str | tuple[str, str] | None = None,
        context: RequestContext | None = None,
        body: bytes | Stream,
    ) -> Response: ...

    @overload
    def options(
        self,
        url: str | URL,
        *,
        headers: Mapping[str, str] | Headers | None = None,
        params: Union[
            str,
            Iterable[Sequence[str]],
            Mapping[str, Union[str, Sequence[str]]],
            URLSearchParams,
        ]
        | None = None,
        auth: str | tuple[str, str] | None = None,
        context: RequestContext | None = None,
        multipart: "Multipart",
    ) -> Response: ...

    @overload
    def options(
        self,
        url: str | URL,
        *,
        headers: Mapping[str, str] | Headers | None = None,
        params: Union[
            str,
            Iterable[Sequence[str]],
            Mapping[str, Union[str, Sequence[str]]],
            URLSearchParams,
        ]
        | None = None,
        auth: str | tuple[str, str] | None = None,
        context: RequestContext | None = None,
    ) -> Response: ...

    def options(
        self,
        url: str | URL,
        *,
        headers: Mapping[str, str] | Headers | None = None,
        params: Union[
            str,
            Iterable[Sequence[str]],
            Mapping[str, Union[str, Sequence[str]]],
            URLSearchParams,
        ]
        | None = None,
        auth: str | tuple[str, str] | None = None,
        context: RequestContext | None = None,
        json: Any | None = None,
        form: Union[
            str,
            Iterable[Sequence[str]],
            Mapping[str, Union[str, Sequence[str]]],
            URLSearchParams,
        ]
        | None = None,
        body: bytes | Stream | None = None,
        multipart: "Multipart | None" = None,
    ) -> Response:
        return self.request(  # type: ignore[call-overload]
            "OPTIONS",
            url,
            headers=headers,
            params=params,
            auth=auth,
            context=context,
            json=json,
            form=form,
            body=body,
            multipart=multipart,
        )

    @overload
    def stream(
        self,
        method: str,
        url: str | URL,
        *,
        headers: Mapping[str, str] | Headers | None = None,
        params: Union[
            str,
            Iterable[Sequence[str]],
            Mapping[str, Union[str, Sequence[str]]],
            URLSearchParams,
        ]
        | None = None,
        auth: str | tuple[str, str] | None = None,
        context: RequestContext | None = None,
        json: Any,
    ) -> AbstractContextManager[Response]: ...

    @overload
    def stream(
        self,
        method: str,
        url: str | URL,
        *,
        headers: Mapping[str, str] | Headers | None = None,
        params: Union[
            str,
            Iterable[Sequence[str]],
            Mapping[str, Union[str, Sequence[str]]],
            URLSearchParams,
        ]
        | None = None,
        auth: str | tuple[str, str] | None = None,
        context: RequestContext | None = None,
        form: Union[
            str,
            Iterable[Sequence[str]],
            Mapping[str, Union[str, Sequence[str]]],
            URLSearchParams,
        ],
    ) -> AbstractContextManager[Response]: ...

    @overload
    def stream(
        self,
        method: str,
        url: str | URL,
        *,
        headers: Mapping[str, str] | Headers | None = None,
        params: Union[
            str,
            Iterable[Sequence[str]],
            Mapping[str, Union[str, Sequence[str]]],
            URLSearchParams,
        ]
        | None = None,
        auth: str | tuple[str, str] | None = None,
        context: RequestContext | None = None,
        body: bytes | Stream,
    ) -> AbstractContextManager[Response]: ...

    @overload
    def stream(
        self,
        method: str,
        url: str | URL,
        *,
        headers: Mapping[str, str] | Headers | None = None,
        params: Union[
            str,
            Iterable[Sequence[str]],
            Mapping[str, Union[str, Sequence[str]]],
            URLSearchParams,
        ]
        | None = None,
        auth: str | tuple[str, str] | None = None,
        context: RequestContext | None = None,
        multipart: "Multipart",
    ) -> AbstractContextManager[Response]: ...

    @overload
    def stream(
        self,
        method: str,
        url: str | URL,
        *,
        headers: Mapping[str, str] | Headers | None = None,
        params: Union[
            str,
            Iterable[Sequence[str]],
            Mapping[str, Union[str, Sequence[str]]],
            URLSearchParams,
        ]
        | None = None,
        auth: str | tuple[str, str] | None = None,
        context: RequestContext | None = None,
    ) -> AbstractContextManager[Response]: ...

    @contextmanager  # type: ignore
    def stream(
        self,
        method: str,
        url: str | URL,
        *,
        headers: Mapping[str, str] | Headers | None = None,
        params: Union[
            str,
            Iterable[Sequence[str]],
            Mapping[str, Union[str, Sequence[str]]],
            URLSearchParams,
        ]
        | None = None,
        auth: str | tuple[str, str] | None = None,
        context: RequestContext | None = None,
        json: Any | None = None,
        form: Union[
            str,
            Iterable[Sequence[str]],
            Mapping[str, Union[str, Sequence[str]]],
            URLSearchParams,
        ]
        | None = None,
        body: bytes | Stream | None = None,
        multipart: "Multipart | None" = None,
    ) -> Iterable[Response]:  # type: ignore
        url_obj = self._merge_url(url, params)
        merged_headers = self._merge_headers(headers, auth=auth)

        if json is not None:
            request = Request(
                method=method,
                url=url_obj,
                headers=merged_headers,
                context=context,
                json=json,
            )
        elif form is not None:
            request = Request(
                method=method,
                url=url_obj,
                headers=merged_headers,
                context=context,
                form=form,
            )
        elif multipart is not None:
            request = Request(
                method=method,
                url=url_obj,
                headers=merged_headers,
                context=context,
                multipart=multipart,
            )
        elif body is not None:
            request = Request(
                method=method,
                url=url_obj,
                headers=merged_headers,
                context=context,
                body=body,
            )
        else:
            request = Request(
                method=method,
                url=url_obj,
                headers=merged_headers,
                context=context,
            )

        response = self.handler.handle(request=request)

        try:
            yield response
        finally:
            response.close()

    def wrap_with_middleware(
        self,
        factory: Callable[
            [BaseHandler],
            BaseMiddleware,
        ],
    ) -> Self:
        self.handler = factory(self.handler)
        return self

    def close(self) -> None:
        handler_to_close: BaseHandler | None = self.handler

        while handler_to_close is not None:
            handler_to_close.close()
            handler_to_close = (
                handler_to_close.next
                if isinstance(
                    handler_to_close,
                    BaseMiddleware,
                )
                else None
            )

    def __enter__(
        self,
    ) -> "Client":
        return self

    def __exit__(
        self,
        exc_type: Any,
        exc_val: Any,
        exc_tb: Any,
    ) -> None:
        self.close()



from __future__ import annotations

import base64
import json
from collections.abc import Iterator
from dataclasses import dataclass
from pathlib import Path
from typing import (
    Any,
    Callable,
    Literal,
    cast,
)
from urllib.parse import (
    parse_qsl,
    urlencode,
    urlsplit,
    urlunsplit,
)

import typing_extensions

from zapros._handlers._async_base import (
    AsyncBaseHandler,
    AsyncBaseMiddleware,
)
from zapros._handlers._common import (
    ensure_async_handler,
    ensure_sync_handler,
)
from zapros._handlers._sync_base import (
    BaseHandler,
    BaseMiddleware,
)
from zapros._models import (
    AsyncClosableStream,
    Headers,
    Request,
    Response,
)

from ..matchers import Matcher


class _CassetteBytesStream:
    def __init__(self, data: bytes) -> None:
        self._data = data
        self._consumed = False

    def __iter__(
        self,
    ) -> Iterator[bytes]:
        return self

    def __next__(self) -> bytes:
        if self._consumed:
            raise StopIteration
        self._consumed = True
        return self._data

    def __aiter__(
        self,
    ) -> _CassetteBytesStream:
        return self

    async def __anext__(self) -> bytes:
        if self._consumed:
            raise StopAsyncIteration
        self._consumed = True
        return self._data


RequestMapper = Callable[[Request], Request]
ResponseMapper = Callable[[Response], Response]
CassetteMode = Literal[
    "all",
    "new_episodes",
    "once",
    "none",
]


class UnhandledRequestError(ValueError):
    pass


def _normalize_url(url: Any) -> str:
    raw = str(url)
    parts = urlsplit(raw)
    normalized_query = urlencode(
        sorted(
            parse_qsl(
                parts.query,
                keep_blank_values=True,
            )
        ),
        doseq=True,
    )
    return urlunsplit(
        (
            parts.scheme,
            parts.netloc,
            parts.path,
            normalized_query,
            parts.fragment,
        )
    )


def _extract_encoding_from_content_type(content_type: str) -> str:
    parts = content_type.split(";")
    for part in parts[1:]:
        stripped = part.strip()
        if stripped.lower().startswith("charset="):
            charset = stripped[8:].strip()
            if charset.startswith('"') and charset.endswith('"'):
                charset = charset[1:-1]
            return charset
    return "utf-8"


def _serialize_body(
    body: bytes | None,
    headers: Headers,
) -> Any:
    if body is None:
        return None

    content_type = headers.get("content-type", "")
    mime_type = content_type.split(";")[0].strip().lower()
    charset = _extract_encoding_from_content_type(content_type)

    if mime_type == "application/json" or mime_type.endswith("+json"):
        return json.loads(body.decode(charset))
    elif mime_type.startswith("text/"):
        return body.decode(charset)
    else:
        return base64.b64encode(body).decode("ascii")


def _deserialize_body(
    body: Any,
    headers: Headers,
) -> bytes | None:
    if body is None:
        return None

    content_type = headers.get("content-type", "")
    mime_type = content_type.split(";")[0].strip().lower()
    charset = _extract_encoding_from_content_type(content_type)

    if isinstance(body, (dict, list)):
        return json.dumps(body, ensure_ascii=False).encode(charset)
    elif isinstance(body, str):
        is_json = mime_type == "application/json" or mime_type.endswith("+json")
        is_text = mime_type.startswith("text/")

        if is_text or is_json:
            return body.encode(charset)
        else:
            return base64.b64decode(body)
    else:
        raise ValueError(f"Unexpected body type: {type(body)}")


def _request_key(
    request: Request,
) -> dict[str, Any]:
    return {
        "method": request.method.upper(),
        "uri": _normalize_url(request.url),
    }


@dataclass
class _StoredInteraction:
    request: dict[str, Any]
    response: dict[str, Any]
    played_back: bool = False


class Modifier:
    def __init__(self, matcher: Matcher) -> None:
        self._matcher = matcher
        self._network_request_mapper: RequestMapper | None = None
        self._network_response_mapper: ResponseMapper | None = None
        self._name: str | None = None

    def matches(self, request: Request) -> bool:
        return self._matcher.match(request)

    def map_network_request(self, fn: RequestMapper) -> Modifier:
        self._network_request_mapper = fn
        return self

    def map_network_response(self, fn: ResponseMapper) -> Modifier:
        self._network_response_mapper = fn
        return self

    def name(self, name: str) -> Modifier:
        self._name = name
        return self

    def apply_network_request(self, request: Request) -> Request:
        if self._network_request_mapper is None:
            return request
        return self._network_request_mapper(request)

    def apply_network_response(self, response: Response) -> Response:
        if self._network_response_mapper is None:
            return response
        return self._network_response_mapper(response)


class Cassette:
    def __init__(
        self,
        *,
        allow_playback_repeats: bool = False,
    ) -> None:
        self._modifiers: list[Modifier] = []
        self._allow_playback_repeats = allow_playback_repeats

    def modifier(self, matcher: Matcher) -> Modifier:
        modifier = Modifier(matcher)
        self._modifiers.append(modifier)
        return modifier

    def prepare_request(self, request: Request) -> Request:

        for modifier in self._modifiers:
            if modifier.matches(request):
                request = modifier.apply_network_request(request)

        return request

    def prepare_network_response(
        self,
        request: Request,
        response: Response,
    ) -> Response:
        current = response

        for modifier in self._modifiers:
            if modifier.matches(request):
                current = modifier.apply_network_response(current)

        return current

    @property
    def allow_playback_repeats(
        self,
    ) -> bool:
        return self._allow_playback_repeats


class CassetteMiddleware(AsyncBaseMiddleware, BaseMiddleware):
    def __init__(
        self,
        cassette: Cassette,
        next_handler: AsyncBaseHandler | BaseHandler,
        *,
        mode: CassetteMode,
        cassette_dir: str,
        cassette_name: str = "default",
    ) -> None:
        self._cassette = cassette
        self._mode = mode
        self.next = cast(BaseHandler, next_handler)
        self.async_next = cast(
            AsyncBaseMiddleware,
            next_handler,
        )
        self._cassette_dir = Path(cassette_dir)
        self._cassette_name = cassette_name
        self._cassette_existed_at_init = self._cassette_path().exists()
        self._interactions = self._load()

    def _cassette_path(self) -> Path:
        return self._cassette_dir / f"{self._cassette_name}.json"

    def _load(
        self,
    ) -> list[_StoredInteraction]:
        path = self._cassette_path()
        if not path.exists():
            return []

        payload = json.loads(path.read_text(encoding="utf-8"))
        return [
            _StoredInteraction(
                request=item["request"],
                response=item["response"],
                played_back=False,  # runtime-only state
            )
            for item in payload
        ]

    def _save(self) -> None:
        self._cassette_dir.mkdir(parents=True, exist_ok=True)
        path = self._cassette_path()
        payload = [
            {
                "request": item.request,
                "response": item.response,
            }
            for item in self._interactions
        ]
        path.write_text(
            json.dumps(
                payload,
                indent=2,
                ensure_ascii=False,
            ),
            encoding="utf-8",
        )

    def _build_replayed_response(self, item: _StoredInteraction) -> Response:
        headers = Headers(item.response["headers"])
        body = _deserialize_body(item.response["body"], headers)
        return Response(
            status=item.response["status"],
            headers=headers,
            content=body,
        )

    def _find_matching_interaction(
        self,
        prepared_request: Request,
    ) -> _StoredInteraction | None:
        key = _request_key(prepared_request)

        if self._cassette.allow_playback_repeats:
            for item in self._interactions:
                if item.request == key:
                    return item
            return None

        for item in self._interactions:
            if item.request == key and not item.played_back:
                return item

        return None

    def _mark_played_back(self, item: _StoredInteraction) -> None:
        if self._cassette.allow_playback_repeats:
            return
        item.played_back = True

    def _can_record(self) -> bool:
        if self._mode == "all":
            return True

        if self._mode == "new_episodes":
            return True

        if self._mode == "once":
            return not self._cassette_existed_at_init

        if self._mode == "none":
            return False

        raise ValueError(f"Unknown cassette mode: {self._mode}")

    def _record_interaction(
        self,
        prepared_request: Request,
        prepared_response: Response,
        response_body: bytes | None,
        response_headers: Headers,
    ) -> None:
        body_content = _serialize_body(
            response_body,
            response_headers,
        )
        self._interactions.append(
            _StoredInteraction(
                request=_request_key(prepared_request),
                response={
                    "status": prepared_response.status,
                    "headers": dict(response_headers),
                    "body": body_content,
                },
            )
        )
        self._save()

    async def _amaterialize_request(self, request: Request) -> Request:
        body = request.body
        if body is None or isinstance(body, bytes):
            return request

        return Request(
            method=request.method,
            url=request.url,
            headers=Headers(dict(request.headers)),
            body=b"".join([chunk async for chunk in body]),  # type: ignore
        )

    def _materialize_request(self, request: Request) -> Request:
        body = request.body
        if body is None or isinstance(body, bytes):
            return request

        return Request(
            method=request.method,
            url=request.url,
            headers=Headers(dict(request.headers)),
            body=b"".join(body),  # type: ignore
        )

    async def ahandle(self, request: Request) -> Response:
        handler = ensure_async_handler(self.async_next)

        replayable_request = await self._amaterialize_request(request)

        if isinstance(request.body, AsyncClosableStream):
            await request.body.aclose()

        prepared_request = self._cassette.prepare_request(replayable_request)

        if self._mode != "all":
            stored = self._find_matching_interaction(prepared_request)
            if stored is not None:
                self._mark_played_back(stored)
                return self._build_replayed_response(stored)

        if not self._can_record():
            raise UnhandledRequestError(
                f"No cassette matched request: {prepared_request.method} {_normalize_url(prepared_request.url)}"
            )

        network_response = await handler.ahandle(replayable_request)

        mapped_response = self._cassette.prepare_network_response(replayable_request, network_response)

        if mapped_response.content is None:
            mapped_body = None
            mapped_headers = Headers(mapped_response.headers)
        else:
            mapped_body = await mapped_response.aread()
            mapped_headers = Headers(
                (k, v) for k, v in mapped_response.headers.items() if k.lower() != "content-encoding"
            )

        await mapped_response.aclose()

        replayable_response = Response(
            status=mapped_response.status,
            headers=Headers(mapped_headers),
            content=mapped_body,
        )

        self._record_interaction(
            prepared_request,
            replayable_response,
            mapped_body,
            mapped_headers,
        )
        return replayable_response

    def handle(self, request: Request) -> Response:
        handler = ensure_sync_handler(self.next)

        replayable_request = self._materialize_request(request)
        prepared_request = self._cassette.prepare_request(replayable_request)

        if self._mode != "all":
            stored = self._find_matching_interaction(prepared_request)
            if stored is not None:
                self._mark_played_back(stored)
                return self._build_replayed_response(stored)

        if not self._can_record():
            raise UnhandledRequestError(
                f"No cassette matched request: {prepared_request.method} {_normalize_url(prepared_request.url)}"
            )

        network_response = handler.handle(replayable_request)

        mapped_response = self._cassette.prepare_network_response(replayable_request, network_response)

        if mapped_response.content is None:
            mapped_body = None
            mapped_headers = Headers(mapped_response.headers)
        else:
            mapped_body = mapped_response.read()
            mapped_headers = Headers(
                (k, v) for k, v in mapped_response.headers.items() if k.lower() != "content-encoding"
            )

        mapped_response.close()

        replayable_response = Response(
            status=mapped_response.status,
            headers=Headers(mapped_headers),
            content=mapped_body,
        )

        self._record_interaction(
            prepared_request,
            replayable_response,
            mapped_body,
            mapped_headers,
        )
        return replayable_response


@typing_extensions.deprecated(
    "CassetteHandler is deprecated, use CassetteMiddleware instead. "
    "The name 'Handler' was misleading as this is a middleware, not a terminal handler."
)
class CassetteHandler(CassetteMiddleware):
    pass



import codecs
import zlib
from typing import (
    Any,
    Iterator,
    Protocol,
)


class DecodingError(Exception):
    pass


class ContentDecoder(Protocol):
    def decode(self, data: bytes) -> bytes:
        raise NotImplementedError()

    def flush(self) -> bytes:
        raise NotImplementedError()


class IdentityDecoder:
    def decode(self, data: bytes) -> bytes:
        return data

    def flush(self) -> bytes:
        return b""


class GZipDecoder:
    def __init__(self) -> None:
        self._decompressor = zlib.decompressobj(zlib.MAX_WBITS | 16)

    def decode(self, data: bytes) -> bytes:
        try:
            return self._decompressor.decompress(data)
        except zlib.error as exc:
            raise DecodingError(f"Failed to decode gzip data: {exc}") from exc

    def flush(self) -> bytes:
        try:
            return self._decompressor.flush()
        except zlib.error as exc:
            raise DecodingError(f"Failed to flush gzip decoder: {exc}") from exc


class DeflateDecoder:
    def __init__(self) -> None:
        self._decompressor = zlib.decompressobj()
        self._first_attempt = True

    def decode(self, data: bytes) -> bytes:
        try:
            return self._decompressor.decompress(data)
        except zlib.error as exc:
            if self._first_attempt:
                self._first_attempt = False
                self._decompressor = zlib.decompressobj(-zlib.MAX_WBITS)
                try:
                    return self._decompressor.decompress(data)
                except zlib.error as exc2:
                    raise DecodingError(f"Failed to decode deflate data: {exc2}") from exc2
            raise DecodingError(f"Failed to decode deflate data: {exc}") from exc

    def flush(self) -> bytes:
        try:
            return self._decompressor.flush()
        except zlib.error as exc:
            raise DecodingError(f"Failed to flush deflate decoder: {exc}") from exc


try:
    import brotli  # type: ignore
except ImportError:
    try:
        import brotlicffi as brotli  # type: ignore
    except ImportError:
        brotli = None


class BrotliDecoder:
    def __init__(self) -> None:
        if brotli is None:
            raise ImportError("BrotliDecoder requires brotli or brotlicffi. Install with: pip install zapros[brotli]")
        self._decompressor: Any = brotli.Decompressor()  # type: ignore

    def decode(self, data: bytes) -> bytes:
        try:
            return self._decompressor.process(data)
        except brotli.error as exc:  # type: ignore
            raise DecodingError(f"Failed to decode brotli data: {exc}") from exc

    def flush(self) -> bytes:
        try:
            if hasattr(
                self._decompressor,
                "flush",
            ):
                return self._decompressor.flush()
            return b""
        except brotli.error as exc:  # type: ignore
            raise DecodingError(f"Failed to flush brotli decoder: {exc}") from exc


try:
    import zstandard  # type: ignore
except ImportError:
    zstandard = None


class ZStandardDecoder:
    def __init__(self) -> None:
        if zstandard is None:
            raise ImportError("ZStandardDecoder requires zstandard. Install with: pip install zapros[zstd]")
        self._decompressor = zstandard.ZstdDecompressor().decompressobj()  # type: ignore

    def decode(self, data: bytes) -> bytes:
        try:
            return self._decompressor.decompress(data)  # type: ignore
        except zstandard.ZstdError as exc:  # type: ignore
            raise DecodingError(f"Failed to decode zstd data: {exc}") from exc

    def flush(self) -> bytes:
        return b""


class MultiDecoder:
    def __init__(
        self,
        decoders: list[ContentDecoder],
    ) -> None:
        self._decoders = decoders

    def decode(self, data: bytes) -> bytes:
        for decoder in reversed(self._decoders):
            data = decoder.decode(data)
        return data

    def flush(self) -> bytes:
        result = b""
        for decoder in reversed(self._decoders):
            flushed = decoder.flush()
            if flushed:
                for dec in self._decoders[self._decoders.index(decoder) + 1 :]:
                    flushed = dec.decode(flushed)
                result += flushed
        return result


class TextDecoder:
    def __init__(self, encoding: str = "utf-8") -> None:
        self.decoder = codecs.getincrementaldecoder(encoding)(errors="replace")

    def decode(self, data: bytes) -> str:
        return self.decoder.decode(data)

    def flush(self) -> str:
        return self.decoder.decode(b"", True)


class ByteChunker:
    def __init__(self, chunk_size: int) -> None:
        self._chunk_size = chunk_size
        self._buffer = bytearray()

    def feed(self, data: bytes) -> Iterator[bytes]:
        self._buffer.extend(data)
        while len(self._buffer) >= self._chunk_size:
            chunk = bytes(self._buffer[: self._chunk_size])
            del self._buffer[: self._chunk_size]
            yield chunk

    def flush(self) -> bytes:
        if self._buffer:
            result = bytes(self._buffer)
            self._buffer.clear()
            return result
        return b""


SUPPORTED_DECODERS: dict[str, type[ContentDecoder]] = {
    "identity": IdentityDecoder,
    "gzip": GZipDecoder,
    "deflate": DeflateDecoder,
    "br": BrotliDecoder,
    "zstd": ZStandardDecoder,
}

_available_encodings: list[str] = []
if zstandard is not None:
    _available_encodings.append("zstd")
if brotli is not None:
    _available_encodings.append("br")
_available_encodings.extend(
    [
        "gzip",
        "deflate",
    ]
)
ACCEPT_ENCODING = ", ".join(_available_encodings)



import re
from typing import Iterator as TIterator

import pytest
from inline_snapshot import snapshot

from zapros._multipart import (
    Multipart,
    MultipartStream,
    Part,
    _escape_quoted_string,
    _generate_part_headers,
)


class _TestSyncStream:
    def __init__(self, data: bytes) -> None:
        self._data = data
        self._consumed = False

    def __iter__(
        self,
    ) -> TIterator[bytes]:
        return self

    def __next__(self) -> bytes:
        if self._consumed:
            raise StopIteration
        self._consumed = True
        return self._data


class _TestAsyncStream:
    def __init__(self, data: bytes) -> None:
        self._data = data
        self._consumed = False

    def __aiter__(
        self,
    ) -> "_TestAsyncStream":
        return self

    async def __anext__(self) -> bytes:
        if self._consumed:
            raise StopAsyncIteration
        self._consumed = True
        return self._data


def test_escape_quoted_string_no_special_chars():
    result = _escape_quoted_string("hello")
    assert result == snapshot("hello")


def test_escape_quoted_string_with_backslash():
    result = _escape_quoted_string("hello\\world")
    assert result == snapshot("hello\\\\world")


def test_escape_quoted_string_with_quotes():
    result = _escape_quoted_string('hello"world')
    assert result == snapshot('hello\\"world')


def test_escape_quoted_string_with_both():
    result = _escape_quoted_string('hello\\"world')
    assert result == snapshot('hello\\\\\\"world')


def test_part_bytes():
    data = b"test content"
    part = Part.bytes(data)

    assert part.content_type == snapshot("application/octet-stream")
    assert part.filename is None
    assert isinstance(part.content, bytes)
    assert part.content == snapshot(b"test content")


def test_part_async_bytes():
    data = b"async content"
    part = Part.bytes(data)

    assert part.content_type == snapshot("application/octet-stream")
    assert part.filename is None


def test_part_stream():
    stream = _TestSyncStream(b"stream content")
    part = Part.stream(stream)

    assert part.content is stream
    assert part.content_type == snapshot("application/octet-stream")
    assert part.filename is None


def test_part_async_stream():
    stream = _TestAsyncStream(b"async stream content")
    part = Part.async_stream(stream)

    assert part.content is stream
    assert part.content_type == snapshot("application/octet-stream")
    assert part.filename is None


def test_part_text():
    text = "hello world"
    part = Part.text(text)

    assert part.content_type == snapshot("text/plain; charset=utf-8")
    assert part.filename is None
    assert isinstance(part.content, bytes)
    assert part.content == snapshot(b"hello world")


def test_part_async_text():
    text = "async hello"
    part = Part.text(text)

    assert part.content_type == snapshot("text/plain; charset=utf-8")
    assert part.filename is None


def test_part_file_name():
    part = Part.bytes(b"data")
    result = part.file_name("test.txt")

    assert result is part
    assert part.filename == snapshot("test.txt")


def test_part_mime_type():
    part = Part.bytes(b"data")
    result = part.mime_type("image/png")

    assert result is part
    assert part.content_type == snapshot("image/png")


def test_part_fluent_api():
    part = Part.bytes(b"image data").file_name("photo.jpg").mime_type("image/jpeg")

    assert part.filename == snapshot("photo.jpg")
    assert part.content_type == snapshot("image/jpeg")


def test_generate_part_headers_no_filename():
    headers = _generate_part_headers("field_name", None, "text/plain")

    assert headers == snapshot(b'Content-Disposition: form-data; name="field_name"\r\nContent-Type: text/plain\r\n\r\n')


def test_generate_part_headers_with_filename():
    headers = _generate_part_headers(
        "file_field",
        "document.pdf",
        "application/pdf",
    )

    assert headers == snapshot(
        b'Content-Disposition: form-data; name="file_field"; filename="document.pdf"\r\n'
        b"Content-Type: application/pdf\r\n\r\n"
    )


def test_generate_part_headers_escaped_name():
    headers = _generate_part_headers('field"name', None, "text/plain")

    assert headers == snapshot(
        b'Content-Disposition: form-data; name="field\\"name"\r\nContent-Type: text/plain\r\n\r\n'
    )


def test_generate_part_headers_escaped_filename():
    headers = _generate_part_headers(
        "field",
        'file"name.txt',
        "text/plain",
    )

    assert headers == snapshot(
        b'Content-Disposition: form-data; name="field"; filename="file\\"name.txt"\r\nContent-Type: text/plain\r\n\r\n'
    )


def test_multipart_stream_single_part():
    part = Part.stream(_TestSyncStream(b"value1"))
    part.content_type = "text/plain; charset=utf-8"
    parts = [("field1", part)]
    boundary = "test-boundary"

    stream = MultipartStream(parts, boundary)
    result = b"".join(stream)

    assert result == snapshot(
        b'--test-boundary\r\nContent-Disposition: form-data; name="field1"\r\n'
        b"Content-Type: text/plain; charset=utf-8\r\n\r\nvalue1\r\n--test-boundary--\r\n"
    )


def test_multipart_stream_multiple_parts():
    part1 = Part.stream(_TestSyncStream(b"value1"))
    part1.content_type = "text/plain; charset=utf-8"
    part2 = Part.stream(_TestSyncStream(b"value2"))
    part2.content_type = "text/plain; charset=utf-8"
    parts = [
        ("field1", part1),
        ("field2", part2),
    ]
    boundary = "test-boundary"

    stream = MultipartStream(parts, boundary)
    result = b"".join(stream)

    assert result == snapshot(
        b'--test-boundary\r\nContent-Disposition: form-data; name="field1"\r\n'
        b"Content-Type: text/plain; charset=utf-8\r\n\r\nvalue1\r\n"
        b'--test-boundary\r\nContent-Disposition: form-data; name="field2"\r\n'
        b"Content-Type: text/plain; charset=utf-8\r\n\r\nvalue2\r\n--test-boundary--\r\n"
    )


def test_multipart_stream_with_file():
    part = Part.stream(_TestSyncStream(b"file content"))
    part.file_name("test.txt").mime_type("text/plain")
    parts = [("file", part)]
    boundary = "boundary123"

    stream = MultipartStream(parts, boundary)
    result = b"".join(stream)

    assert result == snapshot(
        b'--boundary123\r\nContent-Disposition: form-data; name="file"; filename="test.txt"\r\n'
        b"Content-Type: text/plain\r\n\r\nfile content\r\n--boundary123--\r\n"
    )


def test_multipart_stream_close():
    part1 = Part.bytes(b"data1")
    part2 = Part.bytes(b"data2")
    parts = [
        ("field1", part1),
        ("field2", part2),
    ]

    stream = MultipartStream(parts, "boundary")
    stream.close()


def test_multipart_stream_iteration():
    part = Part.stream(_TestSyncStream(b"x" * 100))
    parts = [("field", part)]
    boundary = "b"

    stream = MultipartStream(parts, boundary)
    chunks = list(stream)

    assert len(chunks) > 0
    assert b"".join(chunks).startswith(b"--b\r\n")


@pytest.mark.asyncio
async def test_async_multipart_stream_single_part():
    part = Part.async_stream(_TestAsyncStream(b"async value"))
    part.content_type = "text/plain; charset=utf-8"
    parts = [("field1", part)]
    boundary = "async-boundary"

    stream = MultipartStream(parts, boundary)
    result = b""
    async for chunk in stream:
        result += chunk

    assert result == snapshot(
        b'--async-boundary\r\nContent-Disposition: form-data; name="field1"\r\n'
        b"Content-Type: text/plain; charset=utf-8\r\n\r\nasync value\r\n--async-boundary--\r\n"
    )


@pytest.mark.asyncio
async def test_async_multipart_stream_multiple_parts():
    part1 = Part.async_stream(_TestAsyncStream(b"value1"))
    part1.content_type = "text/plain; charset=utf-8"
    part2 = Part.async_stream(_TestAsyncStream(b"value2"))
    part2.content_type = "text/plain; charset=utf-8"
    parts = [
        ("field1", part1),
        ("field2", part2),
    ]
    boundary = "boundary"

    stream = MultipartStream(parts, boundary)
    result = b""
    async for chunk in stream:
        result += chunk

    assert result == snapshot(
        b'--boundary\r\nContent-Disposition: form-data; name="field1"\r\n'
        b"Content-Type: text/plain; charset=utf-8\r\n\r\nvalue1\r\n"
        b'--boundary\r\nContent-Disposition: form-data; name="field2"\r\n'
        b"Content-Type: text/plain; charset=utf-8\r\n\r\nvalue2\r\n--boundary--\r\n"
    )


@pytest.mark.asyncio
async def test_async_multipart_stream_with_file():
    part = Part.async_stream(_TestAsyncStream(b"binary data"))
    part.file_name("data.bin").mime_type("application/octet-stream")
    parts = [("upload", part)]
    boundary = "file-boundary"

    stream = MultipartStream(parts, boundary)
    result = b""
    async for chunk in stream:
        result += chunk

    assert result == snapshot(
        b'--file-boundary\r\nContent-Disposition: form-data; name="upload"; filename="data.bin"\r\n'
        b"Content-Type: application/octet-stream\r\n\r\nbinary data\r\n--file-boundary--\r\n"
    )


@pytest.mark.asyncio
async def test_async_multipart_stream_aclose():
    part1 = Part.bytes(b"data1")
    part2 = Part.bytes(b"data2")
    parts = [
        ("field1", part1),
        ("field2", part2),
    ]

    stream = MultipartStream(parts, "boundary")
    await stream.aclose()


def test_multipart_default_boundary():
    mp = Multipart()

    assert mp._boundary.startswith("----FormBoundary")
    assert len(mp._boundary) == snapshot(48)


def test_multipart_custom_boundary():
    mp = Multipart(boundary="custom-boundary-123")

    assert mp._boundary == snapshot("custom-boundary-123")


def test_multipart_text():
    mp = Multipart(boundary="test")
    result = mp.text("username", "john_doe")

    assert result is mp
    assert len(mp._parts) == snapshot(1)

    name, part = mp._parts[0]
    assert name == snapshot("username")
    assert part.content_type == snapshot("text/plain; charset=utf-8")


def test_multipart_text_multiple():
    mp = Multipart()
    mp.text("field1", "value1").text("field2", "value2")

    assert len(mp._parts) == snapshot(2)


def test_multipart_part():
    mp = Multipart()
    custom_part = Part.bytes(b"custom").mime_type("custom/type")
    result = mp.part("custom_field", custom_part)

    assert result is mp
    assert len(mp._parts) == snapshot(1)

    name, part = mp._parts[0]
    assert name == snapshot("custom_field")
    assert part is custom_part


def test_multipart_to_stream():
    mp = Multipart(boundary="boundary")
    mp.text("field", "value")

    result = mp.to_body()

    assert isinstance(result, bytes)
    assert b"--boundary\r\n" in result
    assert b"value" in result


def test_multipart_to_async_body():
    mp = Multipart(boundary="boundary")
    mp.text("field", "value")

    result = mp.to_async_body()

    assert isinstance(result, bytes)


@pytest.mark.asyncio
async def test_multipart_to_async_body_iteration():
    mp = Multipart(boundary="async-b")
    mp.part("name", Part.text("test"))

    result = mp.to_async_body()

    assert result == snapshot(
        b'--async-b\r\nContent-Disposition: form-data; name="name"\r\n'
        b"Content-Type: text/plain; charset=utf-8\r\n\r\ntest\r\n--async-b--\r\n"
    )


def test_multipart_content_type():
    mp = Multipart(boundary="my-boundary")

    assert mp.content_type == snapshot('multipart/form-data; boundary="my-boundary"')


def test_multipart_complex_example():
    mp = Multipart(boundary="WebKitFormBoundary")
    mp.text("username", "alice")
    mp.text("email", "alice@example.com")
    mp.part(
        "avatar",
        Part.bytes(b"PNG_DATA").file_name("avatar.png").mime_type("image/png"),
    )

    result = mp.to_body()

    assert result == snapshot(
        b'--WebKitFormBoundary\r\nContent-Disposition: form-data; name="username"\r\n'
        b"Content-Type: text/plain; charset=utf-8\r\n\r\nalice\r\n"
        b'--WebKitFormBoundary\r\nContent-Disposition: form-data; name="email"\r\n'
        b"Content-Type: text/plain; charset=utf-8\r\n\r\nalice@example.com\r\n"
        b'--WebKitFormBoundary\r\nContent-Disposition: form-data; name="avatar"; filename="avatar.png"\r\n'
        b"Content-Type: image/png\r\n\r\nPNG_DATA\r\n--WebKitFormBoundary--\r\n"
    )


def test_multipart_stream_proper_format():
    mp = Multipart(boundary="B")
    mp.text("f", "v")

    result = mp.to_body()

    assert result == snapshot(
        b'--B\r\nContent-Disposition: form-data; name="f"\r\n'
        b"Content-Type: text/plain; charset=utf-8\r\n\r\nv\r\n--B--\r\n"
    )


@pytest.mark.asyncio
async def test_async_multipart_stream_proper_format():
    mp = Multipart(boundary="B")
    mp.part("f", Part.text("v"))

    result = mp.to_async_body()

    assert result == snapshot(
        b'--B\r\nContent-Disposition: form-data; name="f"\r\n'
        b"Content-Type: text/plain; charset=utf-8\r\n\r\nv\r\n--B--\r\n"
    )


def test_part_text_with_unicode():
    text = "Hello 世界 🌍"
    part = Part.text(text)

    assert isinstance(part.content, bytes)
    assert part.content == snapshot(b"Hello \xe4\xb8\x96\xe7\x95\x8c \xf0\x9f\x8c\x8d")
    assert part.content_type == snapshot("text/plain; charset=utf-8")


def test_multipart_text_with_unicode():
    mp = Multipart(boundary="unicode")
    mp.text("message", "Hello 世界")

    result = mp.to_body()

    assert result == snapshot(
        b'--unicode\r\nContent-Disposition: form-data; name="message"\r\n'
        b"Content-Type: text/plain; charset=utf-8\r\n\r\nHello \xe4\xb8\x96\xe7\x95\x8c\r\n--unicode--\r\n"
    )


def test_escape_quoted_string_empty():
    result = _escape_quoted_string("")
    assert result == snapshot("")


def test_multipart_empty():
    mp = Multipart(boundary="empty")
    result = mp.to_body()

    assert result == snapshot(b"--empty--\r\n")


@pytest.mark.asyncio
async def test_async_multipart_empty():
    mp = Multipart(boundary="empty")
    result = mp.to_async_body()

    assert result == snapshot(b"--empty--\r\n")


def test_multipart_stream_assertion_on_wrong_stream_type():
    async_part = Part.async_stream(_TestAsyncStream(b"data"))
    parts = [("field", async_part)]
    stream = MultipartStream(parts, "boundary")

    with pytest.raises(
        AssertionError,
        match="Part content must be a Stream or bytes",
    ):
        list(stream)


@pytest.mark.asyncio
async def test_async_multipart_stream_assertion_on_wrong_stream_type():
    sync_part = Part.stream(_TestSyncStream(b"data"))
    parts = [("field", sync_part)]
    stream = MultipartStream(parts, "boundary")

    with pytest.raises(
        AssertionError,
        match="Part content must be an AsyncStream or bytes",
    ):
        async for _ in stream:
            pass


def test_multipart_stream_close_assertion_on_wrong_stream_type():
    async_part = Part.async_stream(_TestAsyncStream(b"data"))
    parts = [("field", async_part)]
    stream = MultipartStream(parts, "boundary")

    with pytest.raises(
        RuntimeError,
        match=re.escape("Part content is an AsyncStream, use aclose() to close it"),
    ):
        stream.close()


@pytest.mark.asyncio
async def test_async_multipart_stream_aclose_assertion_on_wrong_stream_type():
    sync_part = Part.stream(_TestSyncStream(b"data"))
    parts = [("field", sync_part)]
    stream = MultipartStream(parts, "boundary")

    with pytest.raises(
        RuntimeError,
        match=re.escape("Part content is a Stream, use close() to close it"),
    ):
        await stream.aclose()



from __future__ import annotations

from typing import cast

import typing_extensions
from pywhatwgurl import URL

from zapros._handlers._common import (
    ensure_async_handler,
    ensure_sync_handler,
)
from zapros._handlers._sync_base import (
    BaseHandler,
    BaseMiddleware,
)

from .._errors import TooManyRedirectsError
from .._models import Headers, Request, Response
from ._async_base import (
    AsyncBaseHandler,
    AsyncBaseMiddleware,
)

REDIRECT_STATUS_CODES = {
    301,
    302,
    303,
    307,
    308,
}

METHOD_PRESERVING_CODES = {307, 308}
METHOD_CONVERTING_CODES = {
    301,
    302,
    303,
}

ALWAYS_STRIP_HEADERS = frozenset(
    {
        "authorization",
        "proxy-authorization",
        "cookie",
        "if-match",
        "if-none-match",
        "if-modified-since",
        "if-unmodified-since",
        "if-range",
        "referer",
        "origin",
    }
)

CONTENT_HEADERS_TO_STRIP = frozenset(
    {
        "content-type",
        "content-length",
        "content-encoding",
        "content-language",
        "content-location",
        "transfer-encoding",
        "digest",
    }
)


class RedirectMiddleware(AsyncBaseMiddleware, BaseMiddleware):
    def __init__(
        self,
        next_handler: AsyncBaseHandler | BaseHandler,
        *,
        max_redirects: int = 10,
    ) -> None:
        self.next = cast(BaseHandler, next_handler)
        self.async_next = cast(
            AsyncBaseMiddleware,
            next_handler,
        )
        self._max_redirects = max_redirects

    def _build_redirect_request(
        self,
        original_request: Request,
        location: str,
        status_code: int,
    ) -> Request:
        original_url = original_request.url

        redirect_url = URL(
            location,
            base=original_url.to_string(),
        )

        if status_code == 303:
            new_method = "HEAD" if original_request.method == "HEAD" else "GET"
        elif status_code in (301, 302):
            new_method = "GET" if original_request.method == "POST" else original_request.method
        else:
            new_method = original_request.method

        headers_list = original_request.headers.list()
        new_headers_dict: Headers = Headers()

        method_changed_to_get_or_head = new_method in (
            "GET",
            "HEAD",
        )

        for k, v in headers_list:
            k_lower = k.lower()

            if k_lower in ALWAYS_STRIP_HEADERS:
                continue

            if method_changed_to_get_or_head and k_lower in CONTENT_HEADERS_TO_STRIP:
                continue

            if k_lower == "host":
                continue

            new_headers_dict[k] = v

        if original_request.body is None or method_changed_to_get_or_head:
            return Request(
                redirect_url,
                new_method,
                headers=new_headers_dict,
                context=original_request.context.copy(),
            )
        elif original_request.is_replayable():
            return Request(
                redirect_url,
                new_method,
                headers=new_headers_dict,
                body=original_request.body,
                context=original_request.context.copy(),
            )
        else:
            raise NotImplementedError("Redirect with non-replayable body is not supported")

    async def ahandle(self, request: Request) -> Response:
        handler = ensure_async_handler(self.async_next)

        current_request = request
        redirect_count = 0

        while True:
            response = await handler.ahandle(current_request)

            if response.status not in REDIRECT_STATUS_CODES:
                return response

            location = response.headers.get("Location")
            if location is None:
                return response

            if redirect_count >= self._max_redirects:
                await response.aclose()
                raise TooManyRedirectsError(f"Exceeded maximum number of redirects ({self._max_redirects})")

            current_request = self._build_redirect_request(
                current_request,
                location,
                response.status,
            )
            redirect_count += 1
            await response.aclose()

    def handle(self, request: Request) -> Response:
        handler = ensure_sync_handler(self.next)

        current_request = request
        redirect_count = 0

        while True:
            response = handler.handle(current_request)

            if response.status not in REDIRECT_STATUS_CODES:
                return response

            location = response.headers.get("Location")
            if location is None:
                return response

            if redirect_count >= self._max_redirects:
                response.close()
                raise TooManyRedirectsError(f"Exceeded maximum number of redirects ({self._max_redirects})")

            current_request = self._build_redirect_request(
                current_request,
                location,
                response.status,
            )
            redirect_count += 1
            response.close()


@typing_extensions.deprecated(
    "RedirectHandler is deprecated, use RedirectMiddleware instead. "
    "The name 'Handler' was misleading as this is a middleware, not a terminal handler."
)
class RedirectHandler(RedirectMiddleware):
    pass



import os
from typing import cast

from pywhatwgurl import URL

from zapros._handlers._async_base import AsyncBaseHandler, AsyncBaseMiddleware
from zapros._handlers._common import ensure_async_handler, ensure_sync_handler
from zapros._handlers._sync_base import BaseHandler, BaseMiddleware
from zapros._models import Request, Response


class ProxyMiddleware(BaseMiddleware, AsyncBaseMiddleware):
    def __init__(self, next_handler: AsyncBaseHandler | BaseHandler) -> None:
        self.next = cast(BaseHandler, next_handler)
        self.async_next = cast(
            AsyncBaseHandler,
            next_handler,
        )

    def _extract_proxy_from_env(self, request: Request) -> URL | None:
        # 1. Pick proxy from env (respect scheme)
        scheme = request.url.protocol[:-1].lower()

        proxy_url = (
            os.environ.get(f"{scheme}_proxy")
            or os.environ.get(f"{scheme.upper()}_PROXY")
            or os.environ.get("ALL_PROXY")
            or os.environ.get("all_proxy")
        )

        if not proxy_url:
            return None

        # 2. Respect NO_PROXY
        no_proxy = os.environ.get("NO_PROXY") or os.environ.get("no_proxy")
        if no_proxy:
            host = request.url.host
            for entry in [e.strip() for e in no_proxy.split(",") if e.strip()]:
                if entry == "*" or host == entry or host.endswith(entry.lstrip(".")):
                    return None

        # 3. Parse proxy URL (string → parse)
        parsed = URL(proxy_url)
        return parsed

    def handle(self, request: Request) -> Response:
        proxy_context = request.context.get("network", {}).get("proxy")
        if proxy_context is not None and "url" in proxy_context:
            return ensure_sync_handler(self.next).handle(request)

        if (proxy_from_env := self._extract_proxy_from_env(request)) is not None:
            request.context.setdefault("network", {}).setdefault("proxy", {})["url"] = proxy_from_env

        return Response(200)

    async def ahandle(self, request: Request) -> Response:
        proxy_context = request.context.get("network", {}).get("proxy")
        if proxy_context is not None and "url" in proxy_context:
            return await ensure_async_handler(self.async_next).ahandle(request)

        if (proxy_from_env := self._extract_proxy_from_env(request)) is not None:
            request.context.setdefault("network", {}).setdefault("proxy", {})["url"] = proxy_from_env
        return Response(200)



from __future__ import annotations

from http.client import HTTPMessage
from http.cookiejar import CookieJar
from typing import cast
from urllib.request import (
    Request as UrllibRequest,
)

import typing_extensions

from zapros._handlers._common import (
    ensure_async_handler,
    ensure_sync_handler,
)

from .._models import Request, Response
from ._async_base import (
    AsyncBaseHandler,
    AsyncBaseMiddleware,
)
from ._sync_base import (
    BaseHandler,
    BaseMiddleware,
)


class _MockHTTPResponse:
    def __init__(self, headers: HTTPMessage) -> None:
        self.msg = headers

    def info(self) -> HTTPMessage:
        return self.msg


class CookieMiddleware(AsyncBaseMiddleware, BaseMiddleware):
    def __init__(
        self,
        next_handler: AsyncBaseHandler | BaseHandler,
        jar: CookieJar | None = None,
    ) -> None:
        self.next = cast(BaseHandler, next_handler)
        self.async_next = cast(
            AsyncBaseHandler,
            next_handler,
        )
        self.jar = jar if jar is not None else CookieJar()

    def _process_request(self, request: Request) -> UrllibRequest:
        req = UrllibRequest(
            str(request.url),
            headers=dict(request.headers.list()),
        )
        self.jar.add_cookie_header(req)

        if req.has_header("Cookie"):
            cookie_value = req.get_header("Cookie")
            if cookie_value:
                if "cookie" in request.headers:
                    request.headers.extend(
                        [
                            (
                                "Cookie",
                                cookie_value,
                            )
                        ]
                    )
                else:
                    request.headers.add(
                        "Cookie",
                        cookie_value,
                    )

        return req

    def _process_response(
        self,
        response: Response,
        req: UrllibRequest,
    ) -> Response:
        headers = HTTPMessage()
        for (
            k,
            v,
        ) in response.headers.items():
            headers[k] = v

        mock_response = _MockHTTPResponse(headers)
        self.jar.extract_cookies(mock_response, req)  # type: ignore[arg-type]

        return response

    async def ahandle(self, request: Request) -> Response:
        handler = ensure_async_handler(self.async_next)
        req = self._process_request(request)
        response = await handler.ahandle(request)
        return self._process_response(response, req)

    def handle(self, request: Request) -> Response:
        handler = ensure_sync_handler(self.next)
        req = self._process_request(request)
        response = handler.handle(request)
        return self._process_response(response, req)


@typing_extensions.deprecated(
    "CookieHandler is deprecated, use CookieMiddleware instead. "
    "The name 'Handler' was misleading as this is a middleware, not a terminal handler."
)
class CookieHandler(CookieMiddleware):
    pass



from __future__ import annotations

import asyncio
import ssl
from typing import Any, Callable, TypeVar

from zapros._constants import DEFAULT_READ_SIZE, DEFAULT_SSL_CONTEXT

from .._handlers._exc_map import map_connect_exceptions, map_read_exceptions, map_write_exceptions
from ._base import AsyncBaseNetworkStream, AsyncBaseTransport

T = TypeVar("T")


class _AsyncTLSState:
    def __init__(
        self,
        reader: asyncio.StreamReader,
        writer: asyncio.StreamWriter,
        ssl_object: ssl.SSLObject,
        incoming_bio: ssl.MemoryBIO,
        outgoing_bio: ssl.MemoryBIO,
    ) -> None:
        self.reader = reader
        self.writer = writer
        self.ssl_object = ssl_object
        self.incoming_bio = incoming_bio
        self.outgoing_bio = outgoing_bio
        self.handshake_complete = False


class AsyncIOStream(AsyncBaseNetworkStream):
    def __init__(
        self,
        reader: asyncio.StreamReader,
        writer: asyncio.StreamWriter,
        *,
        ssl_context: ssl.SSLContext | None = None,
        upgrade_ssl_context: ssl.SSLContext | None = None,
    ) -> None:
        self._reader = reader
        self._writer = writer
        self._closed = False
        self._ssl_context = ssl_context or DEFAULT_SSL_CONTEXT
        self._upgrade_ssl_context = upgrade_ssl_context or DEFAULT_SSL_CONTEXT
        self._tls_state: _AsyncTLSState | None = None

    async def _flush_outgoing(self, timeout: float | None = None) -> None:
        assert self._tls_state is not None
        pending = self._tls_state.outgoing_bio.read(DEFAULT_READ_SIZE)
        if pending:
            with map_write_exceptions():
                self._tls_state.writer.write(pending)
                if timeout is None:
                    await self._tls_state.writer.drain()
                else:
                    await asyncio.wait_for(self._tls_state.writer.drain(), timeout)

    async def _pump_incoming(self, timeout: float | None = None) -> None:
        assert self._tls_state is not None
        with map_read_exceptions():
            if timeout is None:
                data = await self._tls_state.reader.read(DEFAULT_READ_SIZE)
            else:
                data = await asyncio.wait_for(self._tls_state.reader.read(n=DEFAULT_READ_SIZE), timeout)

            if data:
                self._tls_state.incoming_bio.write(data)
            else:
                self._tls_state.incoming_bio.write_eof()

    async def _call_sslobject_method(self, func: Callable[..., T], *args: Any) -> T:
        assert self._tls_state is not None
        while True:
            try:
                result = func(*args)
                await self._flush_outgoing()
                return result
            except ssl.SSLWantReadError:
                await self._flush_outgoing()
                await self._pump_incoming()
            except ssl.SSLWantWriteError:
                await self._flush_outgoing()
            except (ssl.SSLError, ssl.SSLEOFError) as e:
                raise ConnectionError(str(e)) from e

    async def read(self, max_bytes: int, timeout: float | None = None) -> bytes:
        if self._tls_state is not None:
            return await self._call_sslobject_method(self._tls_state.ssl_object.read, max_bytes)

        with map_read_exceptions():
            if timeout is None:
                return await self._reader.read(max_bytes)
            return await asyncio.wait_for(self._reader.read(max_bytes), timeout)

    async def write_all(self, data: bytes, timeout: float | None = None) -> int:
        if self._tls_state is not None:
            await self._call_sslobject_method(self._tls_state.ssl_object.write, data)
            return len(data)

        with map_write_exceptions():
            self._writer.write(data)
            if timeout is None:
                await self._writer.drain()
            else:
                await asyncio.wait_for(self._writer.drain(), timeout)
            return len(data)

    async def close(self) -> None:
        if self._closed:
            return
        self._closed = True
        try:
            self._writer.close()
        except Exception:
            pass

    async def start_tls(self, *, server_hostname: str | None = None) -> AsyncBaseNetworkStream:
        assert self._upgrade_ssl_context is not None

        hostname = server_hostname

        incoming = ssl.MemoryBIO()
        outgoing = ssl.MemoryBIO()

        ssl_object = self._upgrade_ssl_context.wrap_bio(incoming, outgoing, server_side=False, server_hostname=hostname)

        self._tls_state = _AsyncTLSState(self._reader, self._writer, ssl_object, incoming, outgoing)

        await self._call_sslobject_method(self._tls_state.ssl_object.do_handshake)
        self._tls_state.handshake_complete = True

        return self


class AsyncIOTransport(AsyncBaseTransport):
    def __init__(
        self,
        *,
        ssl_context: ssl.SSLContext | None = None,
    ) -> None:
        self.ssl_context = DEFAULT_SSL_CONTEXT if ssl_context is None else ssl_context

    async def aconnect(
        self,
        host: str,
        port: int,
        server_hostname: str | None = None,
        tls: bool = False,
        *,
        timeout: float | None = None,
    ) -> AsyncBaseNetworkStream:
        with map_connect_exceptions():
            if timeout is None:
                reader, writer = await asyncio.open_connection(
                    host,
                    port,
                    ssl=self.ssl_context if tls else None,
                    server_hostname=server_hostname if tls else None,
                    happy_eyeballs_delay=0.25,
                    interleave=1,
                )
            else:
                reader, writer = await asyncio.wait_for(
                    asyncio.open_connection(
                        host,
                        port,
                        ssl=self.ssl_context if tls else None,
                        server_hostname=server_hostname if tls else None,
                        happy_eyeballs_delay=0.25,
                        interleave=1,
                        ssl_handshake_timeout=timeout,
                    ),
                    timeout,
                )

        return AsyncIOStream(
            reader,
            writer,
            ssl_context=self.ssl_context,
            upgrade_ssl_context=self.ssl_context,
        )



import json as json_module
from collections.abc import (
    AsyncIterator as ABCAsyncIterator,
    Iterator as ABCIterator,
)
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncIterator,
    Iterable,
    Iterator,
    Mapping,
    MutableMapping,
    Sequence,
    TypeAlias,
    TypedDict,
    TypeVar,
    Union,
    overload,
)

import typing_extensions
from pywhatwgurl import URL, URLSearchParams

from zapros._errors import AsyncSyncMismatchError, ResponseNotRead
from zapros._io._base import AsyncBaseNetworkStream, BaseNetworkStream
from zapros._multidict import (
    CIMultiDict,
)
from zapros._utils import get_host_header_value

if TYPE_CHECKING:
    from ._multipart import Multipart


from ._constants import (
    CHUNK_SIZE,
    USER_AGENT,
)
from ._decoders import (
    ACCEPT_ENCODING,
    SUPPORTED_DECODERS,
    ByteChunker,
    ContentDecoder,
    IdentityDecoder,
    MultiDecoder,
    TextDecoder,
)


class ClosableStream(Iterator[bytes]):
    def close(self) -> None:
        pass


class TimeoutsContext(TypedDict, total=False):
    connect: float
    read: float
    write: float
    total: float


class CachingContext(TypedDict, total=False):
    """
    When set, indicates that the response to this request should be cached with the specified TTL (in seconds).
    """

    ttl: float

    """
    Indicates that the cache entry should be refreshed on access,
    meaning that the TTL will be reset to the original value whenever the cache entry is accessed.
    """
    refresh_ttl_on_access: bool

    """
    When set, indicates that the cache key should include the request body,
    note that zapros will read the body to include it in the cache key,
    so this should only be used with replayable request bodies.
    """
    body_key: str


class ProxyContext(TypedDict, total=False):
    url: str | URL
    server_hostname: str


class NetworkContext(TypedDict, total=False):
    proxy: ProxyContext


class RequestContext(TypedDict, total=False):
    timeouts: TimeoutsContext
    caching: CachingContext
    network: NetworkContext


class ResponseCachingContext(TypedDict, total=False):
    from_cache: bool
    revalidated: bool
    stored: bool
    created_at: float


class ResponseHandoffContext(TypedDict, total=False):
    network_stream: AsyncBaseNetworkStream | BaseNetworkStream


class ResponseContext(TypedDict, total=False):
    caching: ResponseCachingContext
    handoff: ResponseHandoffContext


Stream: TypeAlias = Union[Iterator[bytes], ClosableStream]


class AsyncClosableStream(AsyncIterator[bytes]):
    async def aclose(self) -> None:
        pass


AsyncStream: TypeAlias = Union[
    AsyncIterator[bytes],
    AsyncClosableStream,
]

_T = TypeVar("_T")
_MISSING = object()


class Headers(MutableMapping[str, str]):
    __slots__ = ("_headers",)

    def __init__(
        self,
        headers: Mapping[str, str] | Iterable[tuple[str, str]] | None = None,
    ) -> None:
        self._headers: CIMultiDict[str] = CIMultiDict(headers or ())

    def __getitem__(self, key: str) -> str:
        return self._headers[key]

    def __setitem__(self, key: str, value: str) -> None:
        self._headers[key] = value

    def __delitem__(self, key: str) -> None:
        del self._headers[key]

    def __iter__(self) -> Iterator[str]:
        return iter(self._headers)

    def __len__(self) -> int:
        return len(self._headers)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({list(self._headers.items())!r})"

    def getall(self, key: str) -> list[str]:
        return self._headers.getall(key)

    def add(self, key: str, value: str) -> None:
        self._headers.add(key, value)

    def extend(
        self,
        headers: Mapping[str, str] | Iterable[tuple[str, str]],
    ) -> None:
        self._headers.extend(headers)

    def copy(self) -> "Headers":
        return Headers(self._headers.copy())

    def list(self) -> list[tuple[str, str]]:
        return list(self._headers.items())


class Request:
    __slots__ = (
        "url",
        "method",
        "headers",
        "body",
        "context",
    )

    @overload
    def __init__(
        self,
        url: URL,
        method: str,
        headers: Headers | Mapping[str, str] | None = None,
        *,
        json: Any,
        context: RequestContext | None = None,
    ) -> None: ...

    @overload
    def __init__(
        self,
        url: URL,
        method: str,
        headers: Headers | Mapping[str, str] | None = None,
        *,
        form: Union[
            str,
            Iterable[Sequence[str]],
            Mapping[str, Union[str, Sequence[str]]],
            URLSearchParams,
        ],
        context: RequestContext | None = None,
    ) -> None: ...

    @overload
    def __init__(
        self,
        url: URL,
        method: str,
        headers: Headers | Mapping[str, str] | None = None,
        *,
        body: bytes | Stream | AsyncStream,
        context: RequestContext | None = None,
    ) -> None: ...

    @overload
    def __init__(
        self,
        url: URL,
        method: str,
        headers: Headers | Mapping[str, str] | None = None,
        *,
        multipart: "Multipart",
        context: RequestContext | None = None,
    ) -> None: ...

    @overload
    def __init__(
        self,
        url: URL,
        method: str,
        headers: Headers | Mapping[str, str] | None = None,
        *,
        text: str,
        context: RequestContext | None = None,
    ) -> None: ...

    @overload
    def __init__(
        self,
        url: URL,
        method: str,
        headers: Headers | Mapping[str, str] | None = None,
        *,
        context: RequestContext | None = None,
    ) -> None: ...

    def __init__(
        self,
        url: URL,
        method: str,
        headers: Headers | Mapping[str, str] | None = None,
        *,
        json: Any | None = None,
        form: Union[
            str,
            Iterable[Sequence[str]],
            Mapping[str, Union[str, Sequence[str]]],
            URLSearchParams,
        ]
        | None = None,
        body: bytes | Stream | AsyncStream | None = None,
        multipart: "Multipart | None" = None,
        text: str | None = None,
        context: RequestContext | None = None,
    ) -> None:
        self.url = url
        self.method = method
        self.headers: Headers = headers if isinstance(headers, Headers) else Headers(headers)
        self.context: RequestContext = context if context is not None else {}

        if "host" not in self.headers and url.hostname:
            self.headers.add(
                "Host",
                get_host_header_value(url.hostname, url.protocol[:-1], url.port),
            )

        if "accept" not in self.headers:
            self.headers.add("Accept", "*/*")
        if "user-agent" not in self.headers:
            self.headers.add("User-Agent", USER_AGENT)
        if "accept-encoding" not in self.headers:
            self.headers.add(
                "Accept-Encoding",
                ACCEPT_ENCODING,
            )

        request_body: bytes | Stream | AsyncStream | None = None

        if json is not None:
            request_body = json_module.dumps(
                json,
                separators=(
                    ",",
                    ":",
                ),
                ensure_ascii=False,
            ).encode("utf-8")
            if "Content-Type" not in self.headers:
                self.headers.add(
                    "Content-Type",
                    "application/json",
                )
        elif form is not None:
            request_body = URLSearchParams(form).to_string().encode("utf-8")
            if "Content-Type" not in self.headers:
                self.headers.add(
                    "Content-Type",
                    "application/x-www-form-urlencoded",
                )
        elif multipart is not None:
            request_body = multipart.to_async_body()
            if "Content-Type" not in self.headers:
                self.headers.add(
                    "Content-Type",
                    multipart.content_type,
                )
        elif text is not None:
            request_body = text.encode("utf-8")
            if "Content-Type" not in self.headers:
                self.headers.add(
                    "Content-Type",
                    "text/plain; charset=utf-8",
                )
        elif body is not None:
            request_body = body

        self.body = request_body

        if isinstance(request_body, bytes):
            if "content-length" not in self.headers:
                self.headers.add(
                    "Content-Length",
                    str(len(request_body)),
                )
        elif isinstance(
            request_body,
            (
                ABCIterator,
                ABCAsyncIterator,
            ),
        ):
            if "transfer-encoding" not in self.headers and "content-length" not in self.headers:
                self.headers.add(
                    "Transfer-Encoding",
                    "chunked",
                )

    def __repr__(self) -> str:
        return f"Request(method={self.method!r}, url={str(self.url)!r})"

    def is_replayable(self) -> bool:
        if self.body is None:
            return True
        if isinstance(self.body, bytes):
            return True
        return False


class Response:
    __slots__ = (
        "status",
        "headers",
        "content",
        "_stream_to_close",
        "_decoder",
        "context",
    )

    @overload
    def __init__(
        self,
        status: int,
        headers: Headers | Mapping[str, str] | Iterable[tuple[str, str]] | None = None,
        *,
        content: AsyncStream | Stream | bytes | None,
        context: ResponseContext | None = None,
    ) -> None: ...

    @overload
    def __init__(
        self,
        status: int,
        headers: Headers | Mapping[str, str] | Iterable[tuple[str, str]] | None = None,
        *,
        text: str,
        context: ResponseContext | None = None,
    ) -> None: ...

    @overload
    def __init__(
        self,
        status: int,
        headers: Headers | Mapping[str, str] | Iterable[tuple[str, str]] | None = None,
        *,
        json: Any,
        context: ResponseContext | None = None,
    ) -> None: ...

    @overload
    def __init__(
        self,
        status: int,
        headers: Headers | Mapping[str, str] | Iterable[tuple[str, str]] | None = None,
        *,
        context: ResponseContext | None = None,
    ) -> None: ...

    def __init__(
        self,
        status: int,
        headers: Headers | Mapping[str, str] | Iterable[tuple[str, str]] | None = None,
        *,
        content: AsyncStream | Stream | bytes | None = None,
        text: str | None = None,
        json: Any | None = None,
        context: ResponseContext | None = None,
    ) -> None:
        provided_bodies = sum(value is not None for value in (content, text, json))
        if provided_bodies > 1:
            raise ValueError("Only one of `content`, `body`, `text`, or `json_data` may be provided.")

        self.status = status
        self.headers: Headers = headers if isinstance(headers, Headers) else Headers(headers) if headers else Headers()
        self.content: AsyncStream | Stream | bytes | None = content
        self._stream_to_close: AsyncClosableStream | ClosableStream | None = (
            content if isinstance(content, (AsyncClosableStream, ClosableStream)) else None
        )
        self._decoder: ContentDecoder | None = None
        self.context: ResponseContext = context if context is not None else {}

        if text is not None:
            encoding = self.encoding
            self._set_static_content(
                text.encode(encoding),
                content_type="text/plain; charset=utf-8",
            )
        elif json is not None:
            self._set_static_content(
                json_module.dumps(json).encode("utf-8"),
                content_type="application/json; charset=utf-8",
            )

        if isinstance(self.content, bytes):
            if "content-length" not in self.headers:
                self.headers.add("Content-Length", str(len(self.content)))
        elif isinstance(self.content, (ABCIterator, ABCAsyncIterator)):
            if "transfer-encoding" not in self.headers and "content-length" not in self.headers:
                self.headers.add("Transfer-Encoding", "chunked")

    def _set_static_content(self, data: bytes, *, content_type: str) -> None:
        self.content = data

        self.headers.pop("Content-Encoding", None)

        if "Content-Type" not in self.headers:
            self.headers["Content-Type"] = content_type

    def _get_content_decoder(self) -> ContentDecoder:
        if self._decoder is not None:
            return self._decoder

        encoding_header = self.headers.get("Content-Encoding")
        if not encoding_header:
            self._decoder = IdentityDecoder()
            return self._decoder

        encodings = [enc.strip().lower() for enc in encoding_header.split(",")]
        decoders: list[ContentDecoder] = []

        for encoding in encodings:
            if encoding in SUPPORTED_DECODERS:
                decoders.append(SUPPORTED_DECODERS[encoding]())
            elif encoding and encoding != "identity":
                pass

        if not decoders:
            self._decoder = IdentityDecoder()
        elif len(decoders) == 1:
            self._decoder = decoders[0]
        else:
            self._decoder = MultiDecoder(decoders)

        return self._decoder

    @property
    def encoding(self) -> str:
        content_type = self.headers.get("Content-Type") or ""
        for part in content_type.split(";"):
            part = part.strip()
            if part.lower().startswith("charset="):
                return part[8:].strip().strip('"')
        return "utf-8"

    @property
    def json(self) -> Any:
        """
        Reads the response body (if it has not already been read), decodes it using
        the response encoding, and parses it as JSON.

        Returns the deserialized Python object.
        """
        return json_module.loads(self.text)

    @typing_extensions.deprecated("Use `.aread()` and then `.json` instead")
    async def ajson(self) -> Any:
        """
        Asynchronously reads the response body (if it has not already been read),
        decodes it using the response encoding, and parses it as JSON.

        Returns the deserialized Python object.

        If the body has already been consumed, this method behaves the same as `.json()`,
        since there is no remaining content to read.

        This is a convenience helper for streaming responses, allowing you to read,
        decode, and parse the body in a single call instead of calling `.aread()`
        and then `.json()` manually.
        """
        return json_module.loads(self.text)

    @property
    def text(self) -> str:
        """
        Reads the entire response body (if it has not already been read),
        decodes it using the response encoding, and returns it as a string.
        """
        if self.content is not None and not isinstance(self.content, bytes):
            raise ResponseNotRead(
                "Response body has not been read yet. Call `.read()` or "
                "`.aread()`first to read the body before accessing `.text`.",
            )

        if self.content is None:
            return ""

        return self.content.decode(self.encoding)

    @typing_extensions.deprecated("Use `.aread()` and then `.text` instead")
    async def atext(self) -> str:
        """
        Asynchronously reads the entire response body (if it has not already been read),
        decodes it using the response encoding, and returns it as a string.

        If the body has already been consumed, this method behaves the same as `.text()`,
        since there is no remaining content to read.

        This is a convenience helper for streaming responses, allowing you to read and
        decode the body in a single call instead of calling `.aread()` and then `.text()`
        manually.
        """
        return (await self.aread()).decode(self.encoding)

    def iter_text(
        self,
        chunk_size: int | None = None,
    ) -> Iterator[str]:
        decoder = TextDecoder(encoding=self.encoding)
        for chunk in self.iter_bytes(chunk_size):
            yield decoder.decode(chunk)
        remaining = decoder.flush()
        if remaining:
            yield remaining

    async def async_iter_text(
        self,
        chunk_size: int | None = None,
    ) -> AsyncIterator[str]:
        decoder = TextDecoder(encoding=self.encoding)
        async for chunk in self.async_iter_bytes(chunk_size):
            yield decoder.decode(chunk)
        remaining = decoder.flush()
        if remaining:
            yield remaining

    def iter_bytes(
        self,
        chunk_size: int | None = None,
    ) -> Iterator[bytes]:
        if chunk_size is None:
            chunk_size = CHUNK_SIZE

        if isinstance(self.content, bytes):
            chunker = ByteChunker(chunk_size)
            yield from chunker.feed(self.content)
            remaining = chunker.flush()
            if remaining:
                yield remaining
            return

        if self.content is None:
            self.content = b""
            return

        if not isinstance(self.content, ABCIterator):
            raise AsyncSyncMismatchError(
                "The stream is not synchronous, try using `async_iter_bytes` instead of `iter_bytes`."
            )

        decoder = self._get_content_decoder()
        decoded_chunks: list[bytes] = []

        for raw_chunk in self.content:
            decoded_chunk = decoder.decode(raw_chunk)
            if decoded_chunk:
                decoded_chunks.append(decoded_chunk)

        final_decoded = decoder.flush()
        if final_decoded:
            decoded_chunks.append(final_decoded)

        body = b"".join(decoded_chunks)
        self.content = body

        chunker = ByteChunker(chunk_size)
        yield from chunker.feed(body)
        remaining = chunker.flush()
        if remaining:
            yield remaining

    async def async_iter_bytes(
        self,
        chunk_size: int | None = None,
    ) -> AsyncIterator[bytes]:
        if chunk_size is None:
            chunk_size = CHUNK_SIZE

        if isinstance(self.content, bytes):
            chunker = ByteChunker(chunk_size)
            for chunk in chunker.feed(self.content):
                yield chunk
            remaining = chunker.flush()
            if remaining:
                yield remaining
            return

        if self.content is None:
            self.content = b""
            return

        if not isinstance(self.content, ABCAsyncIterator):
            raise AsyncSyncMismatchError(
                "The stream is not asynchronous, try using `iter_bytes` instead of `async_iter_bytes`."
            )

        decoder = self._get_content_decoder()
        decoded_chunks: list[bytes] = []

        async for raw_chunk in self.content:
            decoded_chunk = decoder.decode(raw_chunk)
            if decoded_chunk:
                decoded_chunks.append(decoded_chunk)

        final_decoded = decoder.flush()
        if final_decoded:
            decoded_chunks.append(final_decoded)

        body = b"".join(decoded_chunks)
        self.content = body

        chunker = ByteChunker(chunk_size)
        for chunk in chunker.feed(body):
            yield chunk
        remaining = chunker.flush()
        if remaining:
            yield remaining

    def iter_raw(
        self,
        chunk_size: int | None = None,
    ) -> Iterator[bytes]:
        if chunk_size is None:
            chunk_size = CHUNK_SIZE

        if isinstance(self.content, bytes):
            chunker = ByteChunker(chunk_size)
            yield from chunker.feed(self.content)
            remaining = chunker.flush()
            if remaining:
                yield remaining
            return

        if self.content is None:
            return

        if isinstance(self.content, ABCAsyncIterator):
            raise AsyncSyncMismatchError(
                "The stream is not synchronous, try using `async_iter_raw` instead of `iter_raw`."
            )

        chunker = ByteChunker(chunk_size)
        for raw_chunk in self.content:
            for chunk in chunker.feed(raw_chunk):
                yield chunk
        remaining = chunker.flush()
        if remaining:
            yield remaining

    async def async_iter_raw(
        self,
        chunk_size: int | None = None,
    ) -> AsyncIterator[bytes]:
        if chunk_size is None:
            chunk_size = CHUNK_SIZE

        if isinstance(self.content, bytes):
            chunker = ByteChunker(chunk_size)
            for chunk in chunker.feed(self.content):
                yield chunk
            remaining = chunker.flush()
            if remaining:
                yield remaining
            return

        if self.content is None:
            return

        if not isinstance(self.content, ABCAsyncIterator):
            raise AsyncSyncMismatchError(
                "The stream is not asynchronous, try using `iter_raw` instead of `async_iter_raw`."
            )

        chunker = ByteChunker(chunk_size)
        async for raw_chunk in self.content:
            for chunk in chunker.feed(raw_chunk):
                yield chunk
        remaining = chunker.flush()
        if remaining:
            yield remaining

    def read(self) -> bytes:
        if not isinstance(self.content, bytes):
            for _ in self.iter_bytes():
                ...
        return self.content  # type: ignore[return-value]

    async def aread(self) -> bytes:
        if not isinstance(self.content, bytes):
            async for _ in self.async_iter_bytes():
                pass
        return self.content  # type: ignore[return-value]

    def close(self) -> None:
        if self._stream_to_close is None:
            return

        if isinstance(self._stream_to_close, AsyncClosableStream):
            raise AsyncSyncMismatchError("The stream is not synchronous, use `aclose()` instead.")

        self._stream_to_close.close()
        self._stream_to_close = None

    async def aclose(self) -> None:
        if self._stream_to_close is None:
            return

        if isinstance(self._stream_to_close, ClosableStream):
            raise AsyncSyncMismatchError("The stream is not asynchronous, use `close()` instead.")

        await self._stream_to_close.aclose()
        self._stream_to_close = None

    def __repr__(self) -> str:
        return f"Response(status={self.status!r})"

    def __enter__(self) -> "Response":
        return self

    def __exit__(
        self,
        exc_type: Any,
        exc_val: Any,
        exc_tb: Any,
    ) -> None:
        self.close()

    async def __aenter__(self) -> "Response":
        return self

    async def __aexit__(
        self,
        exc_type: Any,
        exc_val: Any,
        exc_tb: Any,
    ) -> None:
        await self.aclose()



import ssl
from importlib.metadata import (
    PackageNotFoundError,
    version,
)

CHUNK_SIZE = 16 * 1024
DEFAULT_READ_SIZE = 1024 * 64
DEFAULT_SSL_CONTEXT = ssl.create_default_context()

DEFAULT_PORTS = {
    "http": 80,
    "https": 443,
}

try:
    _version = version("zapros")
except PackageNotFoundError:
    _version = "0.0.0"

USER_AGENT = f"python-zapros/{_version}"



import sqlite3
import time

import pytest
from hishel import (
    FilterPolicy,
    SpecificationPolicy,
    SyncSqliteStorage,
)

from zapros import Client
from zapros._handlers._caching import (
    CacheMiddleware,
)
from zapros._handlers._mock import (
    Mock as ZaprosMock,
    MockMiddleware,
    MockRouter,
)
from zapros._models import Response


def test_cache_miss_then_hit():
    router = MockRouter()
    ZaprosMock().respond(
        Response(
            status=200,
            headers={"Cache-Control": "max-age=3600"},
        )
    ).expect(1).mount(router)

    handler = MockMiddleware(router)
    storage = SyncSqliteStorage(connection=sqlite3.connect(":memory:"))
    caching_handler = CacheMiddleware(handler, storage=storage)

    with Client(handler=caching_handler) as client:
        response1 = client.get(
            "https://example.com/",
        )
        assert response1.status == 200
        caching1 = response1.context.get("caching", {})
        assert caching1.get("from_cache") is False

        response2 = client.get(
            "https://example.com/",
        )
        assert response2.status == 200
        caching2 = response2.context.get("caching", {})
        assert caching2.get("from_cache") is True

        router.verify()


def test_cache_context_fields():
    router = MockRouter()
    ZaprosMock().respond(
        Response(
            status=200,
            headers={"Cache-Control": "max-age=3600"},
        )
    ).once().mount(router)

    handler = MockMiddleware(router)
    storage = SyncSqliteStorage(connection=sqlite3.connect(":memory:"))
    caching_handler = CacheMiddleware(handler, storage=storage)

    with Client(handler=caching_handler) as client:
        response1 = client.get(
            "https://example.com/",
        )
        caching1 = response1.context.get("caching", {})

        assert caching1.get("from_cache") is False
        assert "created_at" in caching1
        assert isinstance(
            caching1.get("created_at"),
            float,
        )

        response2 = client.get(
            "https://example.com/",
        )
        caching2 = response2.context.get("caching", {})

        assert caching2.get("from_cache") is True
        assert "created_at" in caching2


def test_cache_respects_no_cache_directive():
    router = MockRouter()
    ZaprosMock().respond(
        Response(
            status=200,
            headers={"Cache-Control": "no-cache"},
        )
    ).expect(2).mount(router)

    handler = MockMiddleware(router)
    storage = SyncSqliteStorage(connection=sqlite3.connect(":memory:"))
    caching_handler = CacheMiddleware(handler, storage=storage)

    with Client(handler=caching_handler) as client:
        response1 = client.get(
            "https://example.com/",
        )
        assert response1.status == 200

        response2 = client.get(
            "https://example.com/",
        )
        assert response2.status == 200

        router.verify()


@pytest.mark.xfail(reason="investigate hishel's ttl handling")
def test_cache_ttl_in_context():
    router = MockRouter()
    ZaprosMock().respond(
        Response(
            status=200,
            headers={"Cache-Control": "max-age=3600"},
        )
    ).once().mount(router)

    handler = MockMiddleware(router)
    storage = SyncSqliteStorage(connection=sqlite3.connect(":memory:"))
    caching_handler = CacheMiddleware(handler, storage=storage)

    with Client(handler=caching_handler) as client:
        response1 = client.get(
            "https://example.com/",
            context={
                "caching": {
                    "ttl": 2,
                }
            },
        )
        assert response1.status == 200
        caching1 = response1.context.get("caching", {})
        assert caching1.get("from_cache") is False

        time.sleep(1.1)

        ZaprosMock().respond(
            Response(
                status=200,
                headers={"Cache-Control": "max-age=3600"},
            )
        ).once().mount(router)

        response2 = client.get(
            "https://example.com/",
            context={
                "caching": {
                    "ttl": 1.0,
                }
            },
        )
        assert response2.status == 200
        caching2 = response2.context.get("caching", {})
        assert caching2.get("from_cache") is False


def test_cache_body_key():
    router = MockRouter()
    ZaprosMock().respond(
        Response(
            status=200,
            headers={"Cache-Control": "max-age=3600"},
        )
    ).expect(2).mount(router)

    handler = MockMiddleware(router)
    storage = SyncSqliteStorage(connection=sqlite3.connect(":memory:"))
    caching_handler = CacheMiddleware(handler, storage=storage)

    with Client(handler=caching_handler) as client:
        response1 = client.post(
            "https://example.com/",
            body=b"body1",
            context={"caching": {"body_key": "key1"}},
        )
        assert response1.status == 200

        response2 = client.post(
            "https://example.com/",
            body=b"body2",
            context={"caching": {"body_key": "key2"}},
        )
        assert response2.status == 200

        router.verify()


def test_specification_policy():
    router = MockRouter()
    ZaprosMock().respond(
        Response(
            status=200,
            headers={"Cache-Control": "max-age=3600"},
        )
    ).once().mount(router)

    handler = MockMiddleware(router)
    in_memory_storage = SyncSqliteStorage(connection=sqlite3.connect(":memory:"))
    policy = SpecificationPolicy()
    caching_handler = CacheMiddleware(
        handler,
        storage=in_memory_storage,
        policy=policy,
    )

    with Client(handler=caching_handler) as client:
        response1 = client.get(
            "https://example.com/",
        )
        assert response1.status == 200
        caching1 = response1.context.get("caching", {})
        assert caching1.get("from_cache") is False

        response2 = client.get(
            "https://example.com/",
        )
        assert response2.status == 200
        caching2 = response2.context.get("caching", {})
        assert caching2.get("from_cache") is True

        router.verify()


def test_filter_policy():
    router = MockRouter()
    ZaprosMock().respond(Response(status=200, headers={})).expect(1).mount(router)

    handler = MockMiddleware(router)
    in_memory_storage = SyncSqliteStorage(connection=sqlite3.connect(":memory:"))
    policy = FilterPolicy()
    caching_handler = CacheMiddleware(
        handler,
        storage=in_memory_storage,
        policy=policy,
    )

    with Client(handler=caching_handler) as client:
        response1 = client.get(
            "https://example.com/",
        )
        assert response1.status == 200
        caching1 = response1.context.get("caching", {})
        assert caching1.get("from_cache") is False

        response2 = client.get(
            "https://example.com/",
        )
        assert response2.status == 200
        caching2 = response2.context.get("caching", {})
        assert caching2.get("from_cache") is True

        router.verify()


def test_different_urls_not_cached_together():
    router = MockRouter()
    ZaprosMock().respond(
        Response(
            status=200,
            headers={"Cache-Control": "max-age=3600"},
        )
    ).expect(2).mount(router)

    handler = MockMiddleware(router)
    storage = SyncSqliteStorage(connection=sqlite3.connect(":memory:"))
    caching_handler = CacheMiddleware(handler, storage=storage)

    with Client(handler=caching_handler) as client:
        response1 = client.get(
            "https://example.com/path1",
        )
        assert response1.status == 200

        response2 = client.get(
            "https://example.com/path2",
        )
        assert response2.status == 200

        router.verify()


def test_cache_post_request_with_body_key():
    router = MockRouter()
    ZaprosMock().respond(
        Response(
            status=200,
            headers={"Cache-Control": "max-age=3600"},
        )
    ).once().mount(router)

    handler = MockMiddleware(router)
    storage = SyncSqliteStorage(connection=sqlite3.connect(":memory:"))
    caching_handler = CacheMiddleware(
        handler,
        storage=storage,
        policy=FilterPolicy(),
    )

    with Client(handler=caching_handler) as client:
        response1 = client.post(
            "https://example.com/",
            json={"query": "search"},
            context={"caching": {"body_key": "custom_key"}},
        )
        assert response1.status == 200
        caching1 = response1.context.get("caching", {})
        assert caching1.get("from_cache") is False

        response2 = client.post(
            "https://example.com/",
            json={"query": "search"},
            context={"caching": {"body_key": "custom_key"}},
        )
        assert response2.status == 200
        caching2 = response2.context.get("caching", {})
        assert caching2.get("from_cache") is True

        router.verify()



from __future__ import annotations

import time
from dataclasses import dataclass, field
from typing import Protocol

PoolKey = tuple[str, str, int] | tuple[str, str, int, str, str, int]


class PoolConnection(Protocol):
    def close(self) -> None: ...

    @property
    def is_closed(self) -> bool: ...

    def can_reuse(self) -> bool: ...


class AsyncPoolConnection(Protocol):
    async def close(self) -> None: ...

    @property
    def is_closed(self) -> bool: ...

    def can_reuse(self) -> bool: ...


@dataclass
class IdlePoolConnection:
    conn: PoolConnection
    last_used: float = field(default_factory=time.monotonic)


@dataclass
class AsyncIdlePoolConnection:
    conn: AsyncPoolConnection
    last_used: float = field(default_factory=time.monotonic)



from zapros._errors import (
    AsyncSyncMismatchError,
)
from zapros._handlers._async_base import (
    AsyncBaseHandler,
)
from zapros._handlers._sync_base import (
    BaseHandler,
)


def ensure_async_handler(
    handler: AsyncBaseHandler | BaseHandler,
) -> AsyncBaseHandler:
    if isinstance(handler, AsyncBaseHandler):
        return handler
    raise AsyncSyncMismatchError("Handler was expected to be an AsyncBaseHandler, but it is not.")


def ensure_sync_handler(
    handler: AsyncBaseHandler | BaseHandler,
) -> BaseHandler:
    if isinstance(handler, BaseHandler):
        return handler
    raise AsyncSyncMismatchError("Handler was expected to be a BaseHandler, but it is not.")



# pyright: reportUnknownVariableType=false
from __future__ import annotations

import asyncio
from collections.abc import (
    AsyncIterable as ABCAsyncIterable,
)
from typing import Any, AsyncIterator

from .._errors import (
    ConnectionError,
    ConnectTimeoutError,
    ReadTimeoutError,
    WriteTimeoutError,
)
from .._models import (
    AsyncClosableStream,
    Request,
    Response,
)
from ._async_base import (
    AsyncBaseMiddleware,
)


def _deadline_after(
    timeout: float | None,
) -> float | None:
    if timeout is None:
        return None
    return asyncio.get_running_loop().time() + timeout


def _remaining_time(
    deadline: float | None,
) -> float | None:
    if deadline is None:
        return None
    remaining = deadline - asyncio.get_running_loop().time()
    return max(0.0, remaining)


def _combine_timeouts(
    *timeouts: float | None,
) -> float | None:
    values = [t for t in timeouts if t is not None]
    if not values:
        return None
    return min(values)


def _is_async_iterable(
    value: object,
) -> bool:
    return hasattr(value, "__aiter__")


def _js_error_name(
    exc: BaseException,
) -> str | None:
    name = getattr(exc, "name", None)
    if isinstance(name, str):
        return name

    js_error = getattr(exc, "js_error", None)
    if js_error is not None:
        js_name = getattr(js_error, "name", None)
        if isinstance(js_name, str):
            return js_name

    return None


async def _await_with_timeout(
    awaitable: Any,
    timeout: float | None,
) -> Any:
    if timeout is None:
        return await awaitable
    return await asyncio.wait_for(awaitable, timeout=timeout)


class _JSTimeoutAbort:
    def __init__(self, timeout: float | None) -> None:
        self._controller = AbortController.new()  # type: ignore
        self._timeout_id: Any | None = None

        if timeout is not None:
            timeout_ms = max(0, int(timeout * 1000))
            self._timeout_id = setTimeout(
                lambda: self._controller.abort(),  # type: ignore
                timeout_ms,
            )

    @property
    def signal(self) -> Any:
        return self._controller.signal  # type: ignore

    def abort(self) -> None:
        try:
            self._controller.abort()  # type: ignore
        except Exception:
            pass

    def clear(self) -> None:
        if self._timeout_id is not None:
            clearTimeout(self._timeout_id)
            self._timeout_id = None


try:
    from js import (  # type: ignore
        AbortController,
        Object,
        clearTimeout,
        fetch,
        setTimeout,
    )
    from pyodide.ffi import to_js  # type: ignore

    class PyodideAsyncClosableStream(AsyncClosableStream):  # type: ignore[reportRedeclaration]
        def __init__(
            self,
            js_response: Any,
            *,
            read_timeout: float | None = None,
            total_deadline: float | None = None,
            total_abort: _JSTimeoutAbort | None = None,
        ) -> None:
            self._js_response = js_response
            self._reader = js_response.body.getReader() if js_response.body is not None else None
            self._read_timeout = read_timeout
            self._total_deadline = total_deadline
            self._total_abort = total_abort
            self._closed = False

        def __aiter__(
            self,
        ) -> AsyncIterator[bytes]:
            return self

        def _next_read_timeout(
            self,
        ) -> float | None:
            total_remaining = _remaining_time(self._total_deadline)
            if total_remaining == 0:
                return 0.0
            return _combine_timeouts(
                self._read_timeout,
                total_remaining,
            )

        async def __anext__(
            self,
        ) -> bytes:
            if self._closed:
                raise StopAsyncIteration

            if self._reader is None:
                await self.aclose()
                raise StopAsyncIteration

            try:
                result = await _await_with_timeout(
                    self._reader.read(),
                    self._next_read_timeout(),
                )

                if result.done:
                    await self.aclose()
                    raise StopAsyncIteration

                return bytes(result.value.to_py())

            except asyncio.TimeoutError as e:
                await self.aclose()
                raise ReadTimeoutError("Read operation timed out") from e
            except Exception:
                await self.aclose()
                raise

        async def aclose(self) -> None:
            if self._closed:
                return

            self._closed = True

            try:
                if self._reader is not None:
                    await self._reader.cancel()
            except Exception:
                pass
            finally:
                if self._total_abort is not None:
                    self._total_abort.clear()

    class AsyncPyodideHandler(AsyncBaseMiddleware):  # type: ignore[reportRedeclaration]
        def __init__(
            self,
            *,
            total_timeout: float | None = None,
            connect_timeout: float | None = None,
            read_timeout: float | None = None,
            write_timeout: float | None = None,
        ) -> None:
            self.total_timeout = total_timeout
            self.connect_timeout = connect_timeout
            self.read_timeout = read_timeout
            self.write_timeout = write_timeout

        def _resolve_timeouts(
            self,
            request: Request,
        ) -> tuple[
            float | None,
            float | None,
            float | None,
            float | None,
        ]:
            timeouts_context = request.context.get("timeouts", {})

            req_total = timeouts_context.get("total")
            req_connect = timeouts_context.get("connect")
            req_read = timeouts_context.get("read")
            req_write = timeouts_context.get("write")

            total_timeout = req_total if req_total is not None else self.total_timeout
            connect_timeout = req_connect if req_connect is not None else self.connect_timeout
            read_timeout = req_read if req_read is not None else self.read_timeout
            write_timeout = req_write if req_write is not None else self.write_timeout

            return (
                total_timeout,
                connect_timeout,
                read_timeout,
                write_timeout,
            )

        async def _buffer_async_body(
            self,
            body: ABCAsyncIterable[bytes],
            *,
            write_timeout: float | None,
            total_deadline: float | None,
        ) -> bytes:
            chunks: list[bytes] = []
            iterator = body.__aiter__()

            while True:
                next_timeout = _combine_timeouts(
                    write_timeout,
                    _remaining_time(total_deadline),
                )
                try:
                    chunk = await _await_with_timeout(
                        iterator.__anext__(),
                        next_timeout,
                    )
                except StopAsyncIteration:
                    break
                except asyncio.TimeoutError as e:
                    raise WriteTimeoutError("Write operation timed out") from e

                chunks.append(bytes(chunk))

            return b"".join(chunks)

        def _build_headers(self, js_response: Any) -> list[tuple[str, str]]:
            headers_list: list[tuple[str, str]] = []
            entries_iterator = js_response.headers.entries()

            while True:
                entry = entries_iterator.next()
                if entry.done:
                    break
                key, value = entry.value.to_py()
                headers_list.append(
                    (
                        str(key),
                        str(value),
                    )
                )

            return headers_list

        async def ahandle(self, request: Request) -> Response:
            (
                total_timeout,
                connect_timeout,
                read_timeout,
                write_timeout,
            ) = self._resolve_timeouts(request)

            total_deadline = _deadline_after(total_timeout)

            body_js = None
            if request.body is not None:
                if isinstance(
                    request.body,
                    (
                        bytes,
                        bytearray,
                        memoryview,
                    ),
                ):
                    body_js = to_js(bytes(request.body))
                elif _is_async_iterable(request.body):
                    buffered = await self._buffer_async_body(
                        request.body,  # type: ignore
                        write_timeout=write_timeout,
                        total_deadline=total_deadline,
                    )
                    body_js = to_js(buffered)
                else:
                    raise NotImplementedError("Sync iterators are not supported by AsyncPyodideHandler")

            total_remaining = _remaining_time(total_deadline)
            fetch_timeout = _combine_timeouts(
                connect_timeout,
                total_remaining,
            )

            total_abort = _JSTimeoutAbort(total_remaining)
            if fetch_timeout == total_remaining:
                fetch_abort = total_abort
            else:
                fetch_abort = _JSTimeoutAbort(fetch_timeout)

            fetch_options: dict[str, Any] = {
                "method": request.method,
                "headers": Object.fromEntries(  # type: ignore
                    to_js([[k, v] for k, v in request.headers.list()])  # type: ignore
                ),
                "signal": fetch_abort.signal,
            }

            if body_js is not None:
                fetch_options["body"] = body_js

            fetch_options_js = to_js(
                fetch_options,
                dict_converter=Object.fromEntries,  # type: ignore
            )

            try:
                js_response = await fetch(
                    str(request.url),
                    fetch_options_js,
                )  # type: ignore
            except Exception as e:
                fetch_abort.clear()
                if fetch_abort is not total_abort:
                    total_abort.clear()

                error_name = _js_error_name(e)
                if error_name == "AbortError":
                    raise ConnectTimeoutError(f"Network request failed: {e}") from e
                if error_name in (
                    "TypeError",
                    "NetworkError",
                ):
                    raise ConnectionError(f"Network request failed: {e}") from e
                raise
            else:
                if fetch_abort is not total_abort:
                    fetch_abort.clear()

            headers_list = self._build_headers(js_response)
            status = int(js_response.status)  # type: ignore

            stream = PyodideAsyncClosableStream(
                js_response,
                read_timeout=read_timeout,
                total_deadline=total_deadline,
                total_abort=total_abort,
            )

            return Response(
                status=status,
                headers=headers_list,
                content=stream,
            )

        async def aclose(self) -> None:
            pass

except ImportError:
    # If the js module is not available, we are likely not running in Pyodide.
    # In that case, we can define a dummy AsyncPyodideHandler that raises an error if used.

    class PyodideAsyncClosableStream(AsyncClosableStream):
        def __init__(
            self,
            *args: Any,
            **kwargs: Any,
        ) -> None:
            raise RuntimeError("PyodideAsyncClosableStream can only be used in a Pyodide environment")

        def __aiter__(
            self,
        ) -> AsyncIterator[bytes]:
            raise RuntimeError("PyodideAsyncClosableStream can only be used in a Pyodide environment")

        async def __anext__(
            self,
        ) -> bytes:
            raise RuntimeError("PyodideAsyncClosableStream can only be used in a Pyodide environment")

        async def aclose(self) -> None:
            pass

    class AsyncPyodideHandler(AsyncBaseMiddleware):
        async def ahandle(self, request: Request) -> Response:
            raise RuntimeError("AsyncPyodideHandler can only be used in a Pyodide environment")

        async def aclose(self) -> None:
            pass
